-- $Id: invasion-of-gondolin.lua 3325 2018-04-14 23:54:33Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

add_hooks = function()
  -- We should be okay using a local function here, since this won't be
  -- persisted to the savefile.
  local function give_quest(self, data)
    -- Only care about the player.
    if not self.player then return end
    -- Only if we're level 45 or higher.
    if self.level < 45 then return end
    -- Only if we don't have the Maeglin quest.
    if self:hasQuest('invasion-of-gondolin') then return end
    -- Not on quest levels.
    if game.level.data and game.level.data.quest_level then return end
    -- Not while in Mandos.
    if game.short_name == 'halls-of-mandos' then return end

    -- This will trigger the Liron chat.
    self:grantQuest('invasion-of-gondolin')
  end
  -- In case we level up while on a quest level (as one of my characters
  -- did), we also hook into Actor:actBase.
  class:bindHook('Actor:levelupPost', give_quest)
  class:bindHook('Actor:actBase', give_quest)

  class:bindHook('CharacterSheet:dumpLine', function(self, data)
    local q = game.player:hasQuest('invasion-of-gondolin')
    if q then
      if q:isFailed() then
	table.insert(data.lines, [[You abandonded Gondolin when it most needed you, thus causing its destruction.]])
      elseif q:isEnded() then
	return [[You saved Gondolin from destruction.]]
      end
    end
  end)

  -- AI for Maeglin.
  local AI = require 'engine.interface.ActorAI'
  AI:newAI('tome2-maeglin', function(self)
    local dest = game.level:pickSpot{ type='quest', subtype='maeglin-target' }

    if self.x == dest.x and self.y == dest.y then
      -- Oopsie, Maeglin got to Gondolin.
      local Quest = require 'engine.Quest'
      game.log(game.flash.BAD, 'Maeglin has reached Gondolin!  All hope is lost now!')
      -- This will destroy Gondolin for us.
      game.player:setQuestStatus('invasion-of-gondolin', Quest.FAILED)
      -- Make Maeglin "leave" the level.
      game.level:removeEntity(self)
      return true
    end

    -- If near the player, attack; otherwise head for the exit.
    if self:isNear(game.player.x, game.player.y, 2) then
      return self:runAI('dumb_talented_simple')
    else
      return self:moveDirection(dest.x, dest.y)
    end
  end)
end

id = 'invasion-of-gondolin'
name = 'Invasion of Gondolin'
danger_level = 80	-- cf. src/tables.c line 3994
desc = function(self, who)
  if self:isFailed('snubbed-liron') then
    return [[You refused to return to Gondolin and help protect it from Maeglin and Morgoth's armies; as a result, Gondolin was destroyed.]]
  elseif self:isFailed() then
    return [[You failed to prevent Maeglin from leading Morgoth's armies through the secret paths into Gondolin; as a result, Gondolin was destroyed.]]
  elseif self:isEnded() then
    return [[You slew Maegin, ensuring that Morgoth's armies will not find the secret paths into Gondolin.  As a reward, Turgon increased your max life by 150.]]
  elseif self:isCompleted() then
    return [[You have slain Maegin, ensuring that Morgoth's armies will not find the secret paths into Gondolin.  You should return to Gondolin and inform Turgon.]]
  elseif self:isCompleted('turgon') then
    return [[Maeglin is leading Morgoth's army to Gondolin.  You must follow the secret paths out of Gondolin and slay Maeglin before he can lead Morgoth's forces through those paths to the city.]]
  else
    return [[Liron, rider of Tolan has appeared to transport you back to Gondolin.  (Actually, how are you seeing this?)]]
  end
end

-- cf. util.quest_warn_on_exit()
warn_exit_msg = [[If you leave without killing Maeglin, he will lead Morgoth's armies to Gondolin and it will be destroyed.  Proceed anyway?]]

on_grant = function(self, who)
  -- Do Liron chat.
  local Chat = require 'engine.Chat'
  local npc = ui_util.chat_npc('BRONZE_THUNDERLORD', {
    name = [[Liron, rider of Tolan]],
    faction = 'friendlies',
  })
  local chat = Chat.new('invasion-liron', npc, who)
  ui_util.invoke_chat(chat)
end

on_status_change = function(self, who, status, sub)
  if status == self.DONE then
    -- Quest reward is extra HP.
    who.hp_mod = who.hp_mod + 150
    if who.recalcEverything then who:recalcEverything() end
    who.changed = true
  elseif status == self.FAILED then
    -- Destroy Gondolin.
    -- TODO Special handling if we're actually in Gondolin?
    game.state.destroyed_towns['gondolin'] = true
  end
end

on_quest_mon_die = function(self, who, src)
  if who.define_as == 'MAEGLIN_THE_TRAITOR_OF_GONDOLIN' then
    game.log('#YELLOW#You did it!  Gondolin will remain hidden.#LAST#')
    game.player:setQuestStatus(self.id, self.COMPLETED)
  end
end
