-- $Id: timed_effects.lua 3549 2019-01-16 01:21:04Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- TODO:  Temporary effects from src/xtra1.c:calc_bonuses() that we don't
-- implement yet [(X) indicate effects that don't appear to be actually
-- used anywhere]:
--
--   - Water breathing (tim_water_breath) [X]
--   - Magical breathing (tim_magical_breath) [X]
--   - Time resistance (tim_res_time) [X]
--   - Temporary fire aura (tim_fire_aura) [X]
--   - Oppose light and dark (oppose_ld) [X]
--   - Oppose sound and shards (oppose_ss) [X]
--   - Oppose nexus (oppose_nex) [X]
--   - Mental barrier (tim_mental_barrier) [X]
--   - Temporary ESP (tim_esp) [this may be SENSE_MONSTER?]
--
-- TODO:  Temporary effects that don't currently have apply() functions
-- [(*) indicates that it has an on_timeout() function that may obviate
-- apply(); (**) indicates that it is handled manually elsewhere; (C)
-- indicates that it has callbacks that may obviate apply()]:
--
--   - BLIND
--   - CONFUSED
--   - HALLUCINATION
--   - AFRAID (C)
--   - PARALYZED
--   - CUT (*)
--   - WEAPON_PROJECT (C)
--   - ABSORB_SOULS (C)
--   - DISRUPTION_SHIELD (C)
--   - PROBABILITY_TRAVEL
--   - STERILIZE (** cf. spell.try_breed_monster())
--   - DEADLY_AIM (** cf. PlayerCombat:criticalHit())
--   - THUNDERSTORM (*)
--   - POISON_HANDS (C)
--   - PROTECT_EVIL (C)
--   - DRIP_TREAD (C)
--   - SPELLBINDER (C)
--   - INERTIA_CONTROL (*, C)

local Stats = require "mod.class.interface.ActorStats"
local DamageType = require 'engine.DamageType'

newEffect {
  name = 'BLIND',
  desc = 'Blinded',
  type = 'physical',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
    util.juggle_can_see(who)
  end,
  deactivate = function(who, p)
    util.juggle_can_see(who)
  end,
  on_gain = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is blinded!'), '+Blind'
  end,
  on_lose = function(self, p)
    return self:convertTalentMessage(self, '@Target@ can see again'), '-Blind'
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring { 'You are unable to see.' }
  end,
}

newEffect {
  name = 'POISON',
  desc = 'Poisoned',
  src_desc = 'poison',
  type = 'physical',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is poisoned!'), '+Pois'
  end,
  on_lose = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is no longer poisoned.'), '-Pois'
  end,
  on_merge = util.merge_effect_duration,
  copy_on_merge = { 'src' },
  apply = function(who, p)
    who.no_life_regen = true
  end,
  on_timeout = function(self, p, def)
    if not self:hasEffect(self.EFF_INVULNERABILITY) then
      local src = p.src and util.effect_source(def, p.src) or self
      DamageType:get(DamageType.POISON).projector(src, self.x, self.y, DamageType.POISON, {1, no_eff=true})
    end
  end,
  describe = function(who, p)
    local fmt = 'You are poisoned, taking 1 %s damage per turn'
    return fmt:format(util.colorize_damtype(DamageType.POISON)):toTString()
  end,
}

newEffect {
  name = 'CONFUSED',
  desc = 'Confused',
  type = 'mental',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is confused!'), '+Conf'
  end,
  on_lose = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is less confused now.'), '-Conf'
  end,
  on_timeout = function(self, p)
    -- High-level NPCs shake off confusion more quickly.
    -- [cf. process_monster() in src/melee2.c]
    if not self.player and self.level and self.level >= 10 then
      local d = rng.range(1, math.floor(self.level/10))
      if d < p.dur then
	p.dur = p.dur - d
      else
	self:removeEffect(self.EFF_CONFUSED)
      end
    end
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring {
      [[You are confused, preventing you from casting spells or reading scrolls.]]
    }
  end,
  shader = function(def, who, p)
    if not p then return nil end
    local shad, ok = table.get(game, 'posteffects', 'blur', 'shad')
    if ok and shad then
      shad:uniBlur(2)
      return shad
    end
    return nil
  end,
}

newEffect {
  name = 'HALLUCINATION',
  desc = 'Hallucinating',
  type = 'mental',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    game:tweakVisuals()
    return 'Oh, wow!  Everything looks so cosmic now!', '+Hallu'
  end,
  on_lose = function(self, p)
    game:tweakVisuals{ clear_hallu=true }
    return 'You can see clearly again.', '-Hallu'
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring {
      [[Whoa... dude, you're, like, totally hallucinating and stuff...]]
    }
  end,
}

newEffect {
  name = 'AFRAID',
  desc = 'Afraid',
  type = 'mental',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is terrified!'), '+Fear'
  end,
  on_lose = function(self, p)
    local msg = ('@Target@ %s bolder now.'):format(self.player and 'feels' or 'seems')
    return self:convertTalentMessage(self, msg), '-Fear'
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring {
      'You are terrified of monsters around you, and cannot attack them in melee.'
    }
  end,
  callbacks = {
    on_melee_attack_pre = function(src, eff, target, p)
      -- If we're already not attacking, no need to say anything.
      if p.no_attack then return end
      game.logSeen(src, '%s is too afraid to attack %s!', src:getName{def_art=true, check_seen=true}, target:getName{def_art=true, check_seen=true})
      p.no_attack = true
    end,
  },
}

newEffect {
  name = 'PARALYZED',
  desc = 'Paralyzed',
  type = 'mental',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return (p and p.msg or '#Target# is paralyzed!'), '-Move'
  end,
  on_lose = function(self, p)
    return '#Target# can move again.', '+Move'
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring { 'You are unable to move.' }
  end,
}

local stun_msgs = {
  [0] = { '@Target@ is no longer stunned.', nil, '-Stun', '' },
  [1] = { '@Target@ is stunned.', '+Stun',	'-HvyStun', 'Stunned' },
  [2] = { '@Target@ is heavily stunned.', '+HvyStun', '-KnockOut', 'Heavy stun'  },
  [3] = { '@Target@ is knocked out.', '+KnockOut', nil, 'Knocked out' },
}

newEffect {
  name = 'STUN',
  desc = 'Stunned',
  display = function(eff)
    local lvl = math.min(math.floor((eff.dur + 49)/50), 3)
    return stun_msgs[lvl][4]
  end,
  type = 'physical',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
    p.lvl = math.min(math.floor((p.dur + 49)/50), 3)
  end,
  on_gain = function(self, p)
    local lvl = math.min(math.floor((p.dur + 49)/50), 3)
    return self:convertTalentMessage(self, stun_msgs[lvl][1]), stun_msgs[lvl][2]
  end,
  on_lose = function(self, p)
    return self:convertTalentMessage(self, stun_msgs[0][1]), stun_msgs[0][3]
  end,
  on_merge = function(self, cur, p, def)
    if cur.merge == 'no_replace' then return cur end
    local lvl = math.min(math.floor((cur.dur + p.dur + 49)/50), 3)
    util.merge_effect_duration(self, cur, p, def)
    if lvl > cur.lvl then
      cur.lvl = lvl
      game.logSeen(self, self:convertTalentMessage(self, stun_msgs[lvl][1]))
      def.fly(self, stun_msgs[lvl][2])
    end
    return cur
  end,
  on_timeout = function(self, p, def)
    local lvl = math.min(math.floor((p.dur + 48)/50), 3)
    if lvl < p.lvl then
      p.lvl = lvl
      def.fly(self, stun_msgs[lvl][3])
    end
    if not self.player then
      -- NPCs get a "saving throw" vs. stun based on level.
      -- [cf. process_monster() in src/melee2.c]
      local lev = self.level or 0
      if rng.range(0, 4999) <= lev*lev then
	self:removeEffect(self.EFF_STUN)
      end
    end
  end,
  apply = function(who, p)
    local d = p.dur > 50 and 20 or 5
    who:adjustDis('to_h', -d, true, 'stunned')
    who:adjustDis('to_d', -d, true)
  end,
  describe = function(who, p)
    if p > 100 then return tstring{ 'You are knocked out.' } end
    local fmt = 'You are %sstunned, reducing your accuracy and damage by %d.'
    local d = p.dur > 50 and 20 or 5
    local heavy = p.dur > 50 and 'heavily ' or ''
    return fmt:format(heavy, d):toTString()
  end,
  shader = function(def, who, p)
    if not p then return nil end
    local shad, ok = table.get(game, 'posteffects', 'wobbling', 'shad')
    if ok and shad then
      shad:uniWobbling(1)
      return shad
    end
    return nil
  end,
  fly = function(who, msg)
    if game.flyers and who.x and who.y and game.level.map.seens(who.x, who.y) then
      who:addFlyer(20, (rng.range(0, 2) - 1) * 0.5, -3, msg, {255,100,80})
    end
  end,
}

local function cut_level(n)
  return n > 1000 and 7 or
	 n > 200  and 6 or
	 n > 100  and 5 or
	 n > 50   and 4 or
	 n > 25   and 3 or
	 n > 10   and 2 or
	 n > 1    and 1 or 0
end
local function cut_check_scar(self, dur)
  if self:has('SUST_CHR') then return end
  if rng.range(1,1000) < dur or rng.chance(16) then
    game.logSeen(self, self:convertTalentMessage(self, '@Target@ has been horribly scarred.'))
    self:bumpStatDown('CHA')
  end
end
local cut_msgs = {
  [0] = { '', '' },
  [1] = { '@Target@ has been given a graze.', 'Graze' },
  [2] = { '@Target@ has been given a light cut.', 'Light cut' },
  [3] = { '@Target@ has been given a bad cut.', 'Bad cut' },
  [4] = { '@Target@ has been given a nasty cut.', 'Nasty cut' },
  [5] = { '@Target@ has been given a severe cut.', 'Severe cut' },
  [6] = { '@Target@ has been given a deep gash.', 'Deep gash' },
  [7] = { '@Target@ has been given a mortal wound.', 'Mortal wound' },
}

newEffect {
  name = 'CUT',
  desc = 'Bleeding',
  src_desc = 'bleeding',
  display = function(eff)
    return cut_msgs[cut_level(eff.dur)][2]
  end,
  type = 'physical',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
    cut_check_scar(who, p.dur)
  end,
  on_gain = function(self, p)
    return self:convertTalentMessage(self, cut_msgs[cut_level(p.dur)][1]), '+Cut'
  end,
  on_lose = function(self, p)
    return self:convertTalentMessage(self, '@Target@ is no longer bleeding.'), '-Cut'
  end,
  on_merge = function(self, cur, p, def)
    if cur.merge == 'no_replace' then return cur end
    local lvl1 = cut_level(cur.dur)
    util.merge_effect_duration(self, cur, p, def)
    local lvl2 = cut_level(cur.dur)
    if lvl2 > lvl1 then
      game.logSeen(self, self:convertTalentMessage(self, cut_msgs[lvl2]))
      cut_check_scar(self, cur.dur)
    end
    cur.src = p.src or cur.src
    return cur
  end,
  apply = function(who, p)
    who.no_life_regen = true
  end,
  on_timeout = function(self, p, def)
    if not self:hasEffect(self.EFF_INVULNERABILITY) then
      local dam = p.dur > 200 and 3 or p.dur > 100 and 2 or 1
      local src = p.src and util.effect_source(def, p.src) or self
      DamageType:get(DamageType.PHYSICAL).projector(src, self.x, self.y, DamageType.PHYSICAL, dam)
    end
  end,
  describe = function(who, p)
    local fmt = 'You are bleeding, taking %d %s damage per turn.'
    local dam = p.dur > 200 and 3 or p.dur > 100 and 2 or 1
    local damf = util.colorize_dam(DamageType.PHYSICAL)
    return fmt:format(dam, damf):toTString()
  end,
}

newEffect {
  name = 'SPEED',
  desc = 'Hasted',
  display = function(eff)
    return string.format('Hasted[%+d]', eff.speed)
  end,
  type = 'physical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
    if not who.player then
      -- Player:applyTemporaryEffects() handles this for the player.
      who.speed = who.speed + p.speed
    end
  end,
  deactivate = function(who, p)
    if not who.player then
      -- Player:applyTemporaryEffects() handles this for the player.
      who.speed = who.speed - p.speed
    end
  end,
  on_gain = function(self, p)
    local msg = self == game.player and 'You feel yourself moving faster!' or self:convertTalentMessage(self, '@Target@ starts moving faster.')
    return msg, '+Speed'
  end,
  on_lose = function(self, p)
    local msg = self == game.player and 'You feel yourself slow down.' or self:convertTalentMessage(self, '@Target@ slows down.')
    return msg, '-Speed'
  end,
  on_merge = function(self, cur, p, def)
    if cur.merge == 'no_replace' then return cur end
    util.merge_effect_duration(self, cur, p, def)
    if not self.player then
      -- Player:applyTemporaryEffects() handles this for the player.
      self.speed = self.speed - cur.speed + p.speed
    end
    cur.speed = p.speed
    return cur
  end,
  apply = function(who, p)
    who.speed = who.speed + p.speed
  end,
  describe = function(who, p)
    return tstring {
      ('Your speed is temporarily increased by %d.'):format(math.abs(p.speed))
    }
  end,
}

newEffect {
  name = 'SLOW',
  desc = 'Slowed',
  display = function(eff) return 'Slowed[-10]' end,
  type = 'physical',
  status = 'detrimental',
  activate = function(who, p)
    p.merge = nil
    if who ~= game.player then
      -- Player:applyTemporaryEffects() handles this for the player.
      who.speed = who.speed - 10
    end
  end,
  deactivate = function(who, p)
    if who ~= game.player then
      -- Player:applyTemporaryEffects() handles this for the player.
      who.speed = who.speed + 10
    end
  end,
  on_gain = function(self, p)
    local msg = self == game.player and 'You feel yourself moving slower!' or self:convertTalentMessage(self, '@Target@ starts moving slower.')
    return msg, '+Slow'
  end,
  on_lose = function(self, p)
    local msg = self == game.player and 'You feel yourself speed up.' or self:convertTalentMessage(self, '@Target@ speeds up.')
    return msg, '-Slow'
  end,
  on_merge = function(self, cur, p)
    -- Slow is always cumulative.
    cur.dur = cur.dur + p.dur
  end,
  apply = function(who, p)
    who.speed = who.speed - 10
  end,
  describe = function(who, p)
    return tstring {
      'Your speed is temporarily decreased by 10.'
    }
  end,
}

newEffect {
  name = 'GORGED',
  desc = 'Gorged',
  type = 'physical',
  status = 'detrimental',
  decrease = 0, -- Lasts until our food gets low enough.
  no_dur = true, -- Don't show duration in the status display.
  -- Omit on_merge; setEffect() should always do replacement in our case.
  apply = function(who, p)
    who.speed = who.speed - 10
  end,
  describe = function(who, p)
    return tstring {
      'You have eaten far too much, and moving is uncomfortable.  As a result, your speed is decreased by 10 until your stomach has settled.'
    }
  end,
}

newEffect {
  name = 'FAINTING',
  desc = 'Fainting',
  type = 'physical',
  status = 'detrimental',
  decrease = 0, -- Lasts until our food gets high enough.
  no_dur = true, -- Don't show duration in the status display.
  on_timeout = function(self, p)
    if not self:has('PARALYZED') and rng.percent(10) then
      game.logPlayer(self, 'You faint from lack of food.')
      self:runStop('fainted')
      self:restStop('fainted')
      self:setEffect(self.EFF_PARALYZED, 1 + rng.range(1, 5), {merge='combine'})
    end
  end,
  -- Omit on_merge; setEffect() should always do replacement in our case.
  describe = function(who, p)
    return tstring {
      [[You are dizzy from lack of food.  You will have random fainting spells until you get some food into you.]]
    }
  end,
}

newEffect {
  name = 'STARVING',
  desc = 'Starving',
  src_desc = 'starvation',
  type = 'physical',
  status = 'detrimental',
  decrease = 0, -- Lasts until our food gets high enough.
  no_dur = true, -- Don't show duration in the status display.
  on_timeout = function(self, p, def)
    local hit = (util.FOOD.starve - self.food)/10
    if not self:has('INVULNERABILITY') then
      local src = p.src and util.effect_source(def, p.src) or util.effect_source(def, nil, true)
      self:takeHit(hit, src, 'starvation')
    end
  end,
  -- Omit on_merge; setEffect() should always do replacement in our case.
  describe = function(who, p)
    return tstring {
      [[You are starving to death.  You will slowly lose hit points until you get some food into you.]]
    }
  end,
}

newEffect {
  name = 'LIGHT_SPEED',
  desc = 'Light Speed',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel as if time has stopped!', '+LtSpd'
  end,
  on_lose = function(self, p)
    return 'You feel time returning to its normal rate.', '-LtSpd'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    who.speed = who.speed + 50
  end,
  describe = function(who, p)
    return tstring {
      'Your speed is temporarily increased by 50.'
    }
  end,
}

newEffect {
  name = 'WEAPON_PROJECT',
  desc = 'Weeapon Projection',
  -- TODO Maybe include damage type in this?
  display = function(eff) return 'WpnProj' end,
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your weapon starts glowing.', '+WpnPrj'
  end,
  on_lose = function(self, p)
    return 'Your weapon stops glowing.', '-WpnPrj'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  callbacks = {
    on_melee_blow = function(src, eff, target, hit, wpn)
      if not hit or wpn.type ~= 'weapon' then return end
      if eff.radius and eff.radius > 0 then
	spell.ball_spell(src, nil, eff.dt, eff.damage, eff.radius, target.x, target.y, 1)
      else
	spell.hit_spell(src, nil, eff.dt, eff.damage, target.x, target.y, 1)
      end
    end,
  },
  describe = function(who, p)
    local ret = [[Your weapon is imbued with %s, doing %d damage per hit]]
    if p.radius and p.radius > 0 then
      ret = ret .. ([[ in a radius-%d burst around the target.]]):format(p.radius)
    else
      ret = ret .. [[ to the target.]]
    end
    return ret:format(util.colorize_damtype(p.dt), p.damage):toTString()
  end,
}

newEffect {
  name = 'ABSORB_SOULS',
  desc = 'Absorb Souls',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You start absorbing the souls of your foes.', '+AbSoul'
  end,
  on_lose = function(self, p)
    return 'You stop absorbing the souls of your foes.', '-AbSoul'
  end,
  -- make that work.
  -- Omit on_merge(), so that setEffect() does replacement.
  callbacks = {
    on_kill = function(who, p, target, rsrc)
      if rsrc then return end
      if not target:has('UNDEAD') and not target:has('NONLIVING') then
	game.logPlayer('You absorb the life of the dying soul.')
	who:heal(1 + (target.level or 0)/2 + who:getSkillLevel('NECROMANCY, 40'))
      end
    end,
  },
  describe = function(who, p)
    local fmt = [[You are absorbing the souls of the creatures you kill, healing you for at least %d life per soul (increases with Necromancy skill and the level of the monster killed).  Undead and nonliving creatures will not trigger this, for obvious reasons.]]
    local heal = 1 + p:getSkillLevel('NECROMANCY', 40)
    return fmt:format(heal):toTString()
  end,
}

newEffect {
  name = 'DISRUPTION_SHIELD',
  desc = 'Disruption Shield',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel invulnerable!', '+DisShld'
  end,
  on_lose = function(self, p)
    return 'You are more vulnerable.', '-DisShld'
  end,
  on_merge = function(self, cur, p)
    -- No effect if the effect is already active.
  end,
  describe = function(who, p)
    return tstring {
      [[You are surrounded by a mana shield, directing incoming damage to your Mana rather than hit points.  The shield consumes two points of mana for every point of damage absorbed.
      
#{bold}#NOTE:#{normal}#  This shield will break if you attempt to cast any spell, use any talent, attack in melee, or throw or shoot anything]]
    }
  end,
  callbacks = {
    on_melee_attack_pre = function(src, eff, target, p)
      -- Cancel this effect on attack.
      src:removeEffect(src.EFF_DISRUPTION_SHIELD)
    end,
    on_use_talent_pre = function(src, eff, t, slient, dry_run, p)
      -- Cancel this effect on talent use.
      if not dry_run then src:removeEffect(src.EFF_DISRUPTION_SHIELD) end
    end,
    on_take_hit_pre = function(src, p, attacker, pp, wpn)
      -- No mana?  Nothing we can work with.
      if not src.mana then return end
      -- 2 mana absorbs 1 point of damage.
      if src.mana >= pp.damage * 2 then
	src:incMana(-pp.damage * 2)
	pp.damage = 0
      else
	pp.damage = pp.damage - src.mana/2
	src:incMana(-src.mana)
      end
    end,
    priorities = {
      -- Do this before most stuff.
      on_take_hit_pre = -1,
    },
  },
}

newEffect {
  name = 'INVULNERABILITY',
  desc = 'Invulnerability',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return '#00FFFF#Invulnerability!#LAST#', '+Invuln'
  end,
  on_lose = function(self, p)
    return 'The invulnerability wears off.', '-Invuln'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    who:adjustDis('to_a', 100, true)
    -- src/dungeon.c line 1740 uses 40 here, but that's for ten normal-
    -- speed turns
    who.food_regen_mod = (who.food_regen_mod or 0) + 4
  end,
  describe = function(who, p)
    return tstring {
      [[You are surrounded by a nearly impenetrable shield, blocking most damage and increasing your armor by 100.
      
#{bold}#NOTE:#{normal}#  This shield will break if you attempt to cast any spell, use any talent, attack in melee, or throw or shoot anything]]
    }
  end,
  callbacks = {
    on_melee_attack_pre = function(src, eff, target, p)
      -- Cancel this effect on attack.
      src:removeEffect(src.EFF_INVULNERABILITY)
    end,
    on_use_talent_pre = function(src, eff, t, slient, dry_run, p)
      -- Cancel this effect on talent use.
      if not dry_run then src:removeEffect(src.EFF_INVULNERABILITY) end
    end,
    on_take_hit_pre = function(src, p, attacker, pp, wpn)
      -- If no damage, nothing to do.
      if pp.damage <= 0 then return end
      -- Anything under a certain amount must penetrate our invulnerability
      -- to affect us. [cf. src/spells1.c lines 1306-1317]
      if pp.damage < 9000 then
	if rng.chance(13) then
	  game.logPlayer(src, '#YELLOW#The attack penetrates your shield of invulnerability!#LAST#')
	else
	  pp.damage = 0
	end
      end
    end,
    priorities = {
      -- Handle this before Disruption Shield.
      on_take_hit_pre = -2
    },
  },
}

newEffect {
  name = 'PROBABILITY_TRAVEL',
  desc = 'Probability Travel',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel unstable.', '+PrbTrvl'
  end,
  on_lose = function(self, p)
    return 'You are more stable.', '-PrbTrvl'
  end,
  on_merge = util.merge_effect_duration,
  describe = function(who, p)
    return tstring {
      [[You are temporarily immaterial.  When you hit a wall, you will travel through it and instantly appear on the other side of it.  You can also float up and down between dungeon levels at will]],
    }
  end,
}

newEffect {
  name = 'INVISIBLE',
  desc = 'Invisibility',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
    who:resetViewersCanSeeCache()
  end,
  deactivate = function(who, p)
    who:resetViewersCanSeeCache()
  end,
  on_gain = function(self, p)
    if self.player then
      return 'You feel your body fade away.', '+Invis'
    else
      return [[#Target#'s body fades away.]], '+Invis'
    end
  end,
  on_lose = function(self, p)
    if self.player then
      return 'You are no longer invisible.', '-Invis'
    else
      return '#Target# is longer invisible.', '-Invis'
    end
  end,
  on_merge = util.merge_effect_duration,
  copy_if_merge = { 'power' },
  apply = function(who, p)
    who.invisibility = (who.invisibility or 0) + p.power
  end,
  describe = function(who, p)
    return tstring {
      ('You are temporarily invisible (power %d).'):format(p.power)
    }
  end,
}

newEffect {
  name = 'STERILIZE',
  desc = 'Sterilize',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel an anti-sexual aura.', '+Steri'
  end,
  on_lose = function(self, p)
    return 'You no longer feel an anti-sexual aura.', '-Steri'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  describe = function(who, p)
    return tstring {
      [[You are inhibiting reproduction around you.  Monsters nearby will be unable to breed while this effect is active.]]
    }
  end,
}

newEffect {
  name = 'DIVINE_AIM',
  desc = 'Divine Aim',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel very accurate.', '+DivAim'
  end,
  on_lose = function(self, p)
    return 'You are no longer very accurate.', '-DivAim'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who:adjustDis('to_h', 15, true, 'Divine Aim')
    who:adjustDis('to_d', 15, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Tulkas is improving your aim, conferring +15 to-hit and to-damage.]]
    }
  end,
}

-- A duplicate of Divine Aim with different name and description, for
-- Heroic Ballad to use.
newEffect {
  name = 'ACCURATE_STRIKES',
  desc = 'Accurate Strikes',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel very accurate.', '+AcStrk'
  end,
  on_lose = function(self, p)
    return 'You are no longer very accurate.', '-AcStrk'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who:adjustDis('to_h', 15, true, 'Accurate Strikes')
    who:adjustDis('to_d', 15, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your aim is enhanced, conferring +15 to-hit and to-damage.]]
    }
  end,
}

newEffect {
  name = 'DEADLY_AIM',
  desc = 'Deadly Aim',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel extremely accurate.', '+DxAim'
  end,
  on_lose = function(self, p)
    return 'You are suddenly much less accurate.', '-DxAim'
  end,
  decrease = 0,	-- This gets decreased when used.
  -- Omit on_merge(), so that setEffect() does replacement.
  describe = function(who, p)
    return tstring {
      ([[Tulkas is guiding your hand.  Your next %d weapon attacks will be critical hits.]]):format(p.dur)
    }
  end,
}

-- For Scrolls of Monster Confusion.
newEffect {
  name = 'CONFUSE_MONSTER',
  desc = 'Confuse Monster',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your hands begin to glow.', '+CnfMon'
  end,
  on_lose = function(self, p)
    return 'Your hands stop glowing.', '-CnfMon'
  end,
  decrease = 0, -- This gets canceled when used.
  no_dur = true, -- Don't show duration in the status display.
  callbacks = {
    on_melee_blow = function(src, p, target, hit, wpn)
      if hit and wpn.type == 'weapon' then
	src:removeEffect(src.EFF_CONFUSE_MONSTER)
	obj_util.wpn_confuse(src, target)
      end
    end,
  },
  describe = function(who, p)
    return tstring {
      [[Your hands are glowing dull red.  The next monster you hit with a weapon will be confused.]]
    }
  end,
}

newEffect {
  name = 'FEATHER_FALL',
  desc = 'Feather Fall',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel very light.', '+FFall'
  end,
  on_lose = function(self, p)
    return 'You are suddenly heavier.', '-FFall'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'flags', 'FEATHER', true)
  end,
  describe = function(who, p)
    return tstring { [[You are floating just above the ground.]] }
  end,
}

newEffect {
  name = 'FLIGHT',
  desc = 'Flight',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel able to reach the clouds.', '+Fly'
  end,
  on_lose = function(self, p)
    return 'You are suddenly heavier.', '-Fly'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'flags', 'FLY', true)
  end,
  describe = function(who, p)
    return tstring { [[You are flying.]] }
  end,
}

newEffect {
  name = 'REFLECTION',
  desc = 'Reflection',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You start reflecting the world around you.', '+Refl'
  end,
  on_lose = function(self, p)
    return 'You stop reflecting.', '-Refl'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    table.set(who, 'flags', 'REFLECT', true)
  end,
  describe = function(who, p)
    return tstring {
      [[You are surrounded by a mirror shield that reflects bolts and arrows.]]
    }
  end,
}

newEffect {
  name = 'THUNDERSTORM',
  desc = 'Thunderstorm',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'The air around you charges with lightning!', '+Storm'
  end,
  on_lose = function(self, p)
    return 'The air around you discharges.', '-Storm'
  end,
  on_merge = util.merge_effect_duration,
  on_timeout = function(self, p)
    local cands = {}
    for _, a in ipairs(self.fov.actors_dist) do
      -- Only hurt alive hostile monsters that we can see
      if not a.dead and self:reactionToward(a) < 0 and self:canSee(a) then
	table.insert(cands, a)
      end
    end
    local target = #cands > 0 and rng.table(cands)
    if target then
      local dam = rng.dice(p.dmg_dice[1], p.dmg_dice[2])
      local ddam = { dam/3, object=true }
      local t = self:getTalentFromId(self.T_THUNDERSTORM)
      game.logSeen(a, 'Lightning strikes %s.', target)
      spell.hit_spell(self, t, DamageType.ELEC, ddam, target.x, target.y)
      spell.hit_spell(self, t, DamageType.LITE, ddam, target.x, target.y)
      spell.hit_spell(self, t, DamageType.SOUND, ddam, target.x, target.y)
    end
  end,
  describe = function(who, p)
    local ef = util.colorize_damtype(DamageType.ELEC)
    local sf = util.colorize_damtype(DamageType.SOUND)
    local lf = util.colorize_damtype(DamageType.LITE)
    local ret = [[The air around you is charged with electricity.  Each turn a thunder bolt will strike a random monster in sight for %dd%d damage, split evenly between %s, %s and %s.]]
    return tstring {
      ret:format(p.dmg_dice[1], p.dmg_dice[2], ef, sf, lf)
    }
  end,
}

newEffect {
  name = 'MIMIC_FORM',
  desc = 'Mimic Form',
  display = function(eff) return eff.mimic.name end,
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
    if p.mimic.on_gain then p.mimic.on_gain(who, p.mimic) end
  end,
  deactivate = function(who, p)
    if p.mimic.on_lose then p.mimic.on_lose(who, p.mimic) end
  end,
  on_gain = function(self, p)
    return 'You feel your body change.', '+'..p.mimic.name
  end,
  on_lose = function(self, p)
    return 'You are no longer transformed.', '-'..p.mimic.name
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    p.mimic.apply(who, p.level)
  end,
  describe = function(who, p)
    local mf = p.mimic
    local a = mf.name:gmatch('^[AEIOUaeiou]') and 'an' or 'a'
    local desc = tstring {
      ('You are transformed into %s %s, with the following effects:'):format(a, mf.name),
      true
    }
    desc:merge(mf.info(who, p.level))
    return desc
  end,
}

newEffect {
  name = 'LIGHT',
  desc = 'Light',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You suddenly seem brighter!', '+Lite'
  end,
  on_lose = function(self, p)
    return 'You are no longer bright.', '-Lite'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who.lite = (who.lite or 0) + 2
  end,
  describe = function(who, p)
    return tstring {
      [[You are surrounded by light, increasing your light radius by 2.]]
    }
  end,
}

newEffect {
  name = 'POISON_HANDS',
  desc = 'Poison Hands',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your hands are dripping with venom.', '+PsHand'
  end,
  on_lose = function(self, p)
    return 'Your hands stop dripping with venom.', '-PsHand'
  end,
  on_merge = util.merge_effect_duration,
  callbacks = {
    on_wpn_slay_brand = function(who, p, target, wpn, defs)
      -- Add poison brand.
      defs[#defs+1] = { damtype=DamageType.POISON, 3, 6 }
    end,
  },
  describe = function(who, p)
    return tstring {
      [[Your hands are dripping with venom.  Any weapon you attack with will be treated as though it has a poison brand.]]
    }
  end,
}

newEffect {
  name = 'SHIELD',
  desc = 'Shield',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'A mystic shield forms around your body!', '+Shld'
  end,
  on_lose = function(self, p)
    return 'Your mystic shield crumbles away.', '-Shld'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who:adjustDis('to_a', p.power, true)
  end,
  callbacks = {
    on_hit_in_melee = function(src, p, attacker, wpn)
      if p.on_melee_hit and not src.dead and not attacker.dead then
	local auras = p.on_melee_hit
	local name = p.shield_name or 'shield'
	for _, aura in ipairs(#auras == 0 and { auras } or auras) do
	  local dt = DamageType:get(aura.type)
	  local dam = {
	    rng.dice(aura.dice[1], aura.dice[2]),
	    atk_msg = "'s " .. name .. ' hits %s',
	  }
	  dt.projector(src, attacker.x, attacker.y, aura.type, dam)
	  if attacker.dead then return end
	end
      end
    end,
  },
  describe = function(who, p)
    local ret = ([[You are surrounded by a mystic shield, %screasing your AC by %d.]]):format(p.power < 0 and 'de' or 'in', math.abs(p.power))
    if p.on_melee_hit then
      local auras = p.on_melee_hit
      local descs = {}
      for i, aura in ipairs(#auras == 0 and { auras } or auras) do
	local dtf = util.colorize_damtype(aura.type)
	descs[i] = ('%dd%d %s'):format(aura.dice[1], aura.dice[2], dtf)
      end
      descs = table.concatNice(descs, ', ', ' and ')
      ret = ret .. '  In addition, the shield will do ' .. descs ..
	    ' retaliatory damage to monsters that hit you in melee.'
    end
    return ret:toTString()
  end,
}

-- A duplicate of Shield with different name and description (and sans
-- retaliatory damage), for Hobbit Melodies to use.
newEffect {
  name = 'EVASION',
  desc = 'Evasion',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel very evasive!', '+Evade'
  end,
  on_lose = function(self, p)
    return 'You feel less evasive.', '-Evade'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who:adjustDis('to_a', p.power, true)
  end,
  describe = function(who, p)
    local ret = ([[Your ability to dodge melee attacks is enhanced, effectively %screasing your AC by %d]]):format(p.power < 0 and 'de' or 'in', math.abs(p.power))
    return ret:toTString()
  end,
}

newEffect {
  name = 'BLESSED',
  desc = 'Blessed',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel righteous!', '+Bless'
  end,
  on_lose = function(self, p)
    return 'You feel less righteous.', '-Bless'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    who:adjustDis('to_a', 5, true)
    who:adjustDis('to_h', 10, true, 'Blessed')
  end,
  describe = function(who, p)
    return tstring {
      [[You are blessed, conferring +5 AC and +10 to-hit.]]
    }
  end,
}

newEffect {
  name = 'HEROISM',
  desc = 'Heroism',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel like a hero!', '+Hero'
  end,
  on_lose = function(self, p)
    return 'Your heroism wears off.', '-Hero'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    who:adjustDis('to_h', 12, true, 'Heroism')
    table.set(who, 'tmp_resist_dam', DamageType.FEAR, true)
  end,
  describe = function(who, p)
    local ff = util.colorize_damtype(DamageType.FEAR)
    return tstring {
      ([[You feel like a hero, granting +12 to-hit and resistance to %s.]]):format(ff)
    }
  end,
}

newEffect {
  name = 'SUPERHEROISM',
  desc = 'Super-Heroism',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel like a killing machine!', '+SHero'
  end,
  on_lose = function(self, p)
    return 'You feel less berzerk.', '-SHero'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    who:adjustDis('to_h', 24, true, 'Super-Heroism')
    who:adjustDis('to_a', -10, true)
    table.set(who, 'tmp_resist_dam', DamageType.FEAR, true)
  end,
  describe = function(who, p)
    local ff = util.colorize_damtype(DamageType.FEAR)
    return tstring {
      ([[You feel like a killing machine, granting +24 to-hit and resistance to %s, but reducing your AC by 10.]]):format(ff)
    }
  end,
}

newEffect {
  name = 'HOLY_AURA',
  desc = 'Holy Aura',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel a holy aura surround you!', '+Holy'
  end,
  on_lose = function(self, p)
    return 'Your holy aura vanishes.', '-Holy'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    table.set(who, 'flags', 'HOLD_LIFE', true)
    who.adj_luck = (who.adj_luck or 0) + 5
  end,
  describe = function(who, p)
    return tstring {
      [[You are surrounded by a holy aura, granting +5 luck, +1 light radius and resistance to life draining.]]
    }
  end,
}

newEffect {
  name = 'PROTECT_EVIL',
  desc = 'Protection from Evil',
  display = function(eff) return 'Prot.Evil' end,
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel safe from evil!', '+PrEvil'
  end,
  on_lose = function(self, p)
    return 'You no longer feel safe from evil.', '-PrEvil'
  end,
  on_merge = util.merge_effect_duration,
  callbacks = {
    on_take_hit_pre = function(src, p, attacker, pp, wpn)
      -- Only block melee attacks.
      if pp.damage <= 0 or not wpn then return end
      -- Only block attacks from evil attackers.
      if not attacker.has or not attacker:has('EVIL') then return end
      -- Don't block attacks from high-level attackers.
      if src.level < (attacker.level or 0) then return end
      -- Probability based on target level.
      if rng.percent(50 + src.level) then
	game.logPlayer(src, game.flash.GOOD, '%s is repelled', attacker:getName{def_art=true, check_seen=true}:capitalize())
	attacker:learnFlags('flags', 'EVIL')
	pp.damage = 0
      end
    end,
    priorities = {
      -- Handle this before Invulnerability.
      on_take_hit_pre = -3
    },
  },
  -- Omit on_merge(), so that setEffect() does replacement.
  describe = function(who, p)
    local ret = [[You are protected from evil.  An evil monster of level %d or less will be unable to hit you in melee with probability %d%%.]]
    return tstring {
      ret:format(who.level or 0, (who.level or 0) + 50)
    }
  end,
}

newEffect {
  name = 'REGENERATION',
  desc = 'Regeneration',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your body regenerates much more quickly!', '+Regen'
  end,
  on_lose = function(self, p)
    return 'Your body regenerates much more slowly.', '-Regen'
  end,
  on_merge = util.merge_effect_duration,
  copy_if_merge = { 'power' },
  apply = function(who, p)
    who.life_regen_mod = (who.life_regen_mod or 0) + p.power
    -- src/dungeon.c line 1734 uses power/10 here, but that's for ten
    -- normal-speed turns.
    who.food_regen_mod = (who.food_regen_mod or 0) + p.power/100
  end,
  describe = function(who, p)
    return tstring {
      ([[Your regeneration power is increased by %d.]]):format(p.power)
    }
  end,
}

newEffect {
  name = 'WRAITH_FORM',
  desc = 'Wraith Form',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You leave the physical world and turn into a wraith-being!', '+Wraith'
  end,
  on_lose = function(self, p)
    return 'You feel opaque.', '-Wraith'
  end,
  apply = function(who, p)
    -- src/dungeon.c line 1743 uses 30 here, but that's for ten normal-
    -- speed turns
    who.food_regen_mod = (who.food_regen_mod or 0) + 3
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  describe = function(who, p)
    return tstring {
      [[You are as insubstantial as a wraith.  You can pass through most walls.]]
    }
  end,
}

newEffect {
  name = 'SENSE_MONSTER',
  desc = 'Sense Monster',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel your consciousness expand!', '+ESP'
  end,
  on_lose = function(self, p)
    return 'Your consciousness contracts again.', '-ESP'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    table.set(who, 'esp', 'ALL', true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your consciousness has been expanded, conferring temporary ESP for all monster types.]]
    }
  end,
}

newEffect {
  name = 'SEE_INVIS',
  desc = 'See Invisible',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your eyes feel very sensitive!', '+SeeInv'
  end,
  on_lose = function(self, p)
    return 'Your eyes feel less sensitive.', '-SeeInv'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    table.set(who, 'flags', 'SEE_INVIS', true)
  end,
  describe = function(who, p)
    return tstring {
      'You are able to see invisible creatures.'
    }
  end,
}

newEffect {
  name = 'INFRAVISION',
  desc = 'Infravision',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Your eyes begin to tingle!', '+Infra'
  end,
  on_lose = function(self, p)
    return 'Your eyes stop tingling.', '-Infra'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who.infravision = (who.infravision or 0) + 1
  end,
  describe = function(who, p)
    return tstring {
      'Your infravision range is increased by 1.'
    }
  end,
}

newEffect {
  name = 'RES_ACID',
  desc = 'Resist Acid',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel resistant to acid!', '+RAcid'
  end,
  on_lose = function(self, p)
    return 'You feel less resistant to acid.', '-RAcid'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.ACID, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your resilience against acid is increased, reducing acid damage by 2/3.  This is cumulative with acid damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'RES_LIGHTNING',
  desc = 'Resist Lightning',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel resistant to lightning!', '+RLtng'
  end,
  on_lose = function(self, p)
    return 'You feel less resistant to lightning.', '-RLtng'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.ELEC, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your resilience against electricity is increased, reducing lightning damage by 2/3.  This is cumulative with lightning damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'RES_FIRE',
  desc = 'Resist Fire',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel resistant to fire!', '+RFire'
  end,
  on_lose = function(self, p)
    return 'You feel less resistant to fire.', '-RFire'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.FIRE, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your resilience against fire is increased, reducing fire damage by 2/3.  This is cumulative with fire damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'RES_COLD',
  desc = 'Resist Cold',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel resistant to cold!', '+RCold'
  end,
  on_lose = function(self, p)
    return 'You feel less resistant to cold.', '-RCold'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.COLD, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your resilience against cold is increased, reducing cold damage by 2/3.  This is cumulative with cold damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'RES_POISON',
  desc = 'Resist Poison',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel resistant to poison!', '+RPois'
  end,
  on_lose = function(self, p)
    return 'You feel less resistant to poison.', '-RPois'
  end,
  on_merge = util.merge_effect_duration,
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.POISON, true)
  end,
  describe = function(who, p)
    return tstring {
      [[Your resilience against poison is increased, reducing poison damage by 2/3.  This is cumulative with poison damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'RES_CHAOS_CONF',
  desc = 'Resist Conf/Chaos',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You feel protected against raw chaos.', '+RChaos'
  end,
  on_lose = function(self, p)
    return 'You are no longer protected against raw chaos.', '-RChaos'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    table.set(who, 'tmp_resist_dam', DamageType.CHAOS, true)
    table.set(who, 'tmp_resist_dam', DamageType.CONFUSION, true)
  end,
  describe = function(who, p)
    return tstring {
      [[You are resistant to chaos and confusion, reducing chaos damage by 1/7 to 1/2 and confusion damage by 2/7 to 7/12 and protecting you from confusion effects.  This is #{italic}#not#{normal}# cumulative with chaos or confusion damage reduction from worn or wielded equipment.]]
    }
  end,
}

newEffect {
  name = 'DRIP_TREAD',
  desc = 'Dripping Tread',
  display = function(eff) return 'DrpTread' end,
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'You start dripping raw elemental energies.', '+Tread'
  end,
  on_lose = function(self, p)
    return 'You stop dripping raw elemental energies.', '-Tread'
  end,
  decrease = 0,	-- This gets decreased as we walk.
  -- Omit on_merge(); we shouldn't be doing replacement, since re-casting
  -- the spell cancels the effect.
  describe = function(who, p)
    return tstring {
      [[You are dripping raw elemental energies.  The floor behind you will transform to random terrain types as you walk.]]
    }
  end,
  callbacks = {
    on_move = function(who, eff, moved, force, x, y, ox, oy)
      if moved and not force and (x ~= ox or y ~= oy) then
	local new_grid = spell.geomancy.random_floor()
	if new_grid then
	  local Map = require 'mod.class.Map'
	  game.level.map(ox, oy, Map.TERRAIN, new_grid)
	end
	eff.dur = eff.dur - 1
	if eff.dur <= 0 then who:removeEffect(who.T_DRIP_TREAD) end
      end
    end,
  },
}

newEffect {
  name = 'TREE_ROOTS',
  desc = 'Tree Roots',
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    p.merge = nil
  end,
  on_gain = function(self, p)
    return 'Roots dive into the floor from your feet.', '+Roots'
  end,
  on_lose = function(self, p)
    return 'The roots of your feet suddenly vanish.', '-Roots'
  end,
  -- Omit on_merge(), so that setEffect() does replacement.
  apply = function(who, p)
    who:adjustDis('to_a', p.ac, true)
    who.to_d_melee = (who.to_d_melee or 0) + p.dam
    who:removeEffect(who.EFF_STUN)
  end,
  on_timeout = function(self, p)
    who:removeEffect(who.EFF_STUN)
  end,
  describe = function(who, p)
    local ret = [[Your feet are anchored to the floor by magical tree roots, making you more stable (+%d AC) and able to make stronger attacks (+%d to-damage), but preventing you from moving or teleporting.  You also recover from stun effects almost immediately.]]
    return ret:format(p.ac or 0, p.dam or 0):toTString()
  end,
}

newEffect {
  name = 'SPELLBINDER',
  desc = 'Spellbinder',
  display = function(eff) return 'SpellBnd' end,
  type = 'magical',
  status = 'beneficial',
  decrease = 0, -- Lasts until triggered or canceled.
  no_dur = true, -- Don't show duration in the status display.
  -- Omit on_merge(); we can't re-cast the spell while the effect is
  -- active, so no replacement or merging is possible.
  describe = function(who, p)
    local names = {}
    for _, tid in ipairs(p.tids) do
      local t = who:getTalentFromId(tid)
      if t then names[#names+1] = '• ' .. t.name end
    end
    names = table.concat(names, '\n  ')
    local ret = [[You have bound the following spells to trigger automatically if your life drops below %d%% of maximum:
  ]] .. names
    return ret:format(p.pct)
  end,
  callbacks = {
    on_take_hit_post = function(who, p, src, dam)
      if (who.life * 100)/who.max_life < p.pct then
	game.logPlayer(who, 'Your spellbinder is triggered!')
	who:removeEffect(who.EFF_SPELLBINDER, true)
	for _, tid in ipairs(p.tids) do
	  who:forceUseTalent(tid, { ignore_energy=true, ignore_ressources=true, force_no_fail=true })
	end
      end
    end,
  },
}

newEffect {
  name = 'INERTIA_CONTROL',
  desc = 'Inertia Control',
  display = function(eff) return 'InerCtrl' end,
  type = 'magical',
  status = 'beneficial',
  activate = function(who, p)
    local t = who:getTalentFromId(p.t_id)
    p.timeout = t.inertia[2]
  end,
  decrease = 0,	-- Lasts until canceled.
  no_dur = true, -- Don't show duratin in the status display,
  -- Omit on_merge(); re-casting the spell cancels the effect.
  on_timeout = function(who, p)
    local t = who:getTalentFromId(p.t_id)
    p.timeout = p.timeout - 1
    if p.timeout <= 0 then
      if not game.zone.worldmap then
	who:forceUseTalent(t, { ignore_energy=true, ignore_ressources=true })
      end
      p.timeout = t.inertia[2]
    end
  end,
  describe = function(who, p)
    local t = who:getTalentFromId(p.t_id)
    local mana = spell.mana_cost(who, t)
    local s = p.timeout == 1 and '' or 's'
    local ret = [[You have bound the energy flow of your %s spell so that it casts automatically every %d normal-speed turns (next cast in %d turn%s).  While the binding is in effect, your maximum mana is reduced by %d.]]
    return ret:format(t.name, t.inertia[2], p.timeout, s, mana*4)
  end,
  callbacks = {
    on_player_calc_max_mana = function(who, p)
      local t = who:getTalentFromId(p.t_id)
      local mana = spell.mana_cost(who, t)
      who.max_mana = who.max_mana - 4*mana
    end,
  },
}

newEffect {
  name = 'RECALL',
  desc = 'Recall',
  type = 'magical',
  status = 'beneficial',
  cancel_on_level_change = true,
  activate = function(who, p)
    game.logSeen(who, 'The air around you becomes charged...')
    p.merge = nil
  end,
  deactivate = function(who, p)
    if p.dur > 0 then
      game.logSeen(who, 'A tension leaves the air around you...')
    else
      game:onTickEnd(function()
	-- TODO Check again that we can recall out (hey, the player could
	-- have cast Timed Roots after reading their ?oWoR, you never
	-- know...).
	-- Can't recall out of quest levels.
	if game.level.data and game.level.data.quest_level then
	  game.logPlayer(who, 'The recall is canceled by a powerful magic force!')
	  return
	end
	game.logPlayer(who, 'You feel yourself yanked %s!', p.dest.to_town and 'upwards' or 'downwards')
	game:changeLevel(p.dest.level, p.dest.zone, { recall=true })
      end)
    end
  end,
  -- Omit on_merge(); spell.set_recall() will cancel us if recall is re-cast.
  -- Don't need an apply() method
  describe = function(who, p)
    local ret = [[You are waiting to be recalled back to %s.]]
    return ret:format(p.dest.name or '<<unknown>>'):toTString()
  end,
}

newEffect {
  name = 'WILD_AMBUSH',
  desc = 'Ambush!',
  display = function(eff) return 'Ambush' end,
  type = 'other',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave trap-detected zone.
  no_dur = true, -- Don't show duration in the status display.
  activate = function(who, p)
    game.logPlayer(who, 'You are ambushed!')
  end,
  deactivate = function(who, p)
    game.logPlayer(who, 'You have escaped the ambush!')
  end,
  describe = function(who, p)
    local fmt =  [[You have been ambushed in the wild.  You cannot return to the worldmap via '%s' until you have escaped past the edge of this wilderness area into an adjacent area.]]
    return tstring { fmt:format(util.getFormattedKeyBinds('CHANGE_LEVEL', "' or l'")) }
  end,
}

newEffect {
  name = 'DET_TRAPS',
  desc = 'Detected Traps',
  display = function(eff) return 'DetTrap' end,
  type = 'magical',
  status = 'beneficial',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave trap-detected zone.
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_gain = function(self, p)
    if p.silent_gain then
      p.silent_gain = nil
      return nil, nil
    end
    return 'You enter a trap-detected zone.', '+DetTrap'
  end,
  on_lose = function(self, p)
    return 'You leave a trap-detected zone.', '-DetTrap'
  end,
  describe = function(who, p)
    return tstring { [[You are in an area that you have scanned for traps.]] }
  end,
}

newEffect {
  name = 'DROWNING',
  desc = 'Drowning',
  src_desc = 'drowning',
  display = function(eff) return 'Drown' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  apply = function(who, p)
    who.no_life_regen = true
  end,
  on_timeout = function(who, p, def)
    -- If the player has shed some weight or gained a useful flag, cancel
    -- the effect.
    if who:has('FEATHER') or who:has('WATER_BREATH') or who:has('MAGIC_BREATH') or who:totalCarriedWeight() <= who:carryingCapacity()/2 then
      who:removeEffect(who.EFF_DROWNING)
      return
    end
    local src = p.src and util.effect_source(def, p.src) or util.effect_source(def, nil, true)
    who:takeHit((who.level or 1), src, 'drowning')
  end,
  on_gain = function(self, p)
    return 'You are drowning!', '+Drown'
  end,
  on_lose = function(self, p)
    return 'You are no longer drowning.', '-Drown'
  end,
  describe = function(who, p)
    local fmt = [[You are drowning.  While you remain in deep water while carrying a heavy load (more than half your carrying capacity), you will take %d damage per turn (increases with level).]]
    return tstring { fmt:format(who.level or 1) }
  end,
}

newEffect {
  name = 'ON_SHALLOW_LAVA',
  desc = 'Shallow Lava',
  display = function(eff) return 'ShalLava' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_timeout = function(who, p)
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.FIRE
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, who.level or 1)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local ff = util.colorize_damtype(DamageType.FIRE)
    local fmt = [[You are standing on shallow lava.  While you remain, you will take %d %s damage per turn (increases with level).]]
    return tstring { fmt:format(who.level or 1, ff) }
  end,
}

newEffect {
  name = 'ON_DEEP_LAVA',
  desc = 'Deep Lava',
  display = function(eff) return 'DeepLava' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_timeout = function(who, p)
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.FIRE
    local dam = rng.dice(who.level or 1, 2)
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, dam)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local ff = util.colorize_damtype(DamageType.FIRE)
    local fmt = [[You are standing on deep lava.  While you remain, you will take %dd2 %s damage per turn (increases with level).]]
    return tstring { fmt:format(who.level or 1, ff) }
  end,
}

newEffect {
  name = 'ON_ICE',
  desc = 'On Ice',
  display = function(eff) return 'OnIce' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_gain = function(who, p)
    p.turn_count = 0
  end,
  on_timeout = function(who, p)
    local hit = p.turn_count == 0
    p.turn_count = (p.turn_count + 1) % 50
    if not hit then return end
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.ICE
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, 1)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local cf = util.colorize_damtype(DamageType.ICE)
    local fmt = [[You are standing on ice.  While you remain, you will take 1 %s damage every 50 turns.]]
    return tstring { fmt:format(cf) }
  end,
}

newEffect {
  name = 'ON_NETHER_MIST',
  desc = 'Nether Mist',
  display = function(eff) return 'NethMist' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_gain = function(who, p)
    p.turn_count = 0
  end,
  on_timeout = function(who, p)
    local hit = p.turn_count == 0
    p.turn_count = (p.turn_count + 1) % 40
    if not hit then return end
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.NETHER
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, 1)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local nf = util.colorize_damtype(DamageType.NETHER)
    local fmt = [[You are standing in nether mist.  While you remain, you will take 1 %s damage every 40 turns.]]
    return tstring { fmt:format(nf) }
  end,
}

newEffect {
  name = 'ON_GREAT_FIRE',
  desc = 'Great Fire',
  display = function(eff) return 'GrtFire' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_timeout = function(who, p)
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.HELL_FIRE
    local dam = rng.dice(150, 2)
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, dam)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local hf = util.colorize_damtype(DamageType.HELL_FIRE)
    local fmt = [[You are standing in the Great Fire.  While you remain, you will take 150d2 %s damage per turn.]]
    return tstring { fmt:format(hf) }
  end,
}

newEffect {
  name = 'ON_A_BLAZING_FIRE',
  desc = 'Blazing Fire',
  display = function(eff) return 'BlazFire' end,
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  on_timeout = function(who, p)
    local Map = require 'mod.class.Map'
    local g = game.level.map(who.x, who.y, Map.TERRAIN)
    local dt = DamageType.FIRE
    local dam = rng.dice(who.level or 1, 2)
    local proj = spell.set_project { grid=false, object=false }
    DamageType:get(dt).projector(g, who.x, who.y, dt, dam)
    spell.set_project(proj)
  end,
  describe = function(who, p)
    local ff = util.colorize_damtype(DamageType.FIRE)
    local fmt = [[You are standing in a blazing fire.  While you remain, you will take %dd2 %s damage per turn (increases with level).]]
    return tstring { fmt:format(who.level or 1, ff) }
  end,
}

newEffect {
  name = 'DETECTED',
  desc = 'Detected',
  display = function(eff) return 'Detected' end,
  type = 'magical',
  status = 'other',
  decrease = 0, -- Juggled by Player:stepOneTurnFOV().
  no_dur = true, -- Don't show duration in the status display.
  describe = function(who, p)
    return tstring { [[Target has been magically detected by the player.]] }
  end,
}

newEffect {
  name = 'BLOOD_OF_LIFE',
  desc = 'Blood of Life',
  display = function(eff) return 'BldLife' end,
  type = 'magical',
  status = 'beneficial',
  decrease = 0, -- Lasts until it saves our proverbial.
  no_dur = true, -- Don't show duration in the status display.
  on_gain = function(self, p)
    return 'You feel the blood of life running through your veins!', '+BldLife'
  end,
  on_lose = function(self, p)
    return 'You have been saved by the Blood of Life!', '-BldLife'
  end,
  callbacks = {
    on_die_pre = function(who, eff, src, p)
      -- If something else already saved our proverbial, no need to do
      -- anything.
      if p.no_die then return end
      -- Our work here... is done.
      p.no_die = true
      who:removeEffect(who.EFF_BLOOD_OF_LIFE)
    end,
  },
  describe = function(who, p)
    return tstring { [[The Blood of Life flows through your veins.  It will save you from death — #{italic}#once#{normal}#.]] }
  end,
}

newEffect {
  name = 'BLACK_BREATH',
  desc = 'Black Breath',
  display = function(eff) return 'BlkBrth' end,
  type = 'magical',
  status = 'detrimental',
  decrease = 0, -- Lasts until it's cured by something.
  no_dur = true, -- Don't show duration in the status display.
  equip_flags = { 'flags', 'BLACK_BREATH', sticky=true },
  on_gain = function(self, p)
    p.last_warn_check = game.turn
    -- Slight Hack(TM):  Getting Black Breath from a DG_CURSE displays a
    -- different on-gain message.
    local dg_hack = p.dg_hack
    p.dg_hack = nil
    local msg = dg_hack and 'You feel the coldness of the Black Breath attacking you!' or 'You feel the Black Breath slowly draining you of life...'
    return msg, '+BlkBrth'
  end,
  on_lose = function(self, p)
    return 'The hold of the Black Breath on you is broken!', '-BlkBrth'
  end,
  callbacks = {
    on_global_turn = function(who, eff)
      -- If we're not wearing anything that confers Black Breath, warn
      -- about it. [cf. src/dungeon.c lines 2789-2817]
      if game.turn >= eff.last_warn_check + 3000 then
	eff.last_warn_check = game.turn
	local function bb_cb(o, _, inven)
	  return inven.worn and o:hasFlags('convey_flags', 'BLACK_BREATH')
	end
	if not util.findInAllInventoriesBy(who, bb_cb) then
	  game.logPlayer(who, 'The Black Breath saps your soul!')
	  if who.restStop then
	    who:restStop('Black Breath')
	    who:runStop('Black Breath')
	  end
	end
      end

      -- Drain exp and stats. [cf. src/dungeon.c lines 2878-2893]
      local pct = who:hasFlag('flags', 'RESIST_BLACK_BREATH') and 2 or 5
      if rng.percent(pct) and who.exp > 0 then
	local exp = 1 + math.floor(who.level/5)
	who:gainExp(-exp)
	who.max_exp = who.max_exp - exp
	local Stats = require 'mod.class.interface.ActorStats'
	who:bumpStatDown(rng.table(Stats.stats))
      end
    end,
  },
  describe = function(who, p)
    return tstring {
      [[You are afflicted with the Black Breath.  This malady will slowly drain you of life and leave you as a shadow creature, unless you can find some way to cure yourself.]]
    }
  end,
}

newEffect {
  name = 'MOLD_AFFINITY',
  desc = 'Mold Affinity',
  display = function(eff) return 'Mold Aff' end,
  type = 'magical',
  status = 'beneficial',
  decrease = 0, -- Meant to be permanent
  no_dur = true, -- Don't show duration in the status display.
  on_gain = function(self, p)
    return 'You feel a sudden affinity for mold.', '+MoldAff'
  end,
  on_lose = function(self, p)
    return 'You feel a sudden dislike for mold.', '-MoldAff'
  end,
  apply = function(who, p)
    who:accumTargetTalent('T_GROW_MOLD')
  end,
  describe = function(who, p)
    return tstring {
      [[You have an affinity for mold, and can cause them to grow near you.]]
    }
  end,
}

newEffect {
  name = 'DRAIN_EXP',
  desc = 'Drain Experience',
  display = function(eff) return 'Dr.Exp' end,
  type = 'magical',
  status = 'detrimental',
  decrease = 0, -- Removed when we take off the equipment with the flag.
  no_dur = true, -- Don't show duration in the status display.
  equip_flags = { 'flags', 'DRAIN_EXP' },
  callbacks = {
    on_global_turn = function(who, eff)
      -- 10% chance to drain 1exp every global turn
      -- [cf. src/dungeon.c lines 2962-2970]
      if who.exp and who.exp > 0 and rng.percent(10) then
	who:gainExp(-1)
	who.max_exp = math.max(who.max_exp - 1, who.exp)
      end
    end,
  },
  describe = function(who, p)
    return tstring {
      [[Your experience is being drained.  Every normal-speed turn, you have a 10% chance of non-restorably losing 1 exp.]]
    }
  end,
}

newEffect {
  name = 'DRAIN_MANA',
  desc = 'Drain Mana',
  display = function(eff) return 'Dr.Mana' end,
  type = 'magical',
  status = 'detrimental',
  decrease = 0, -- Removed when we take off the equipment with the flag.
  no_dur = true, -- Don't show duration in the status display.
  equip_flags = { 'drain_mana', power=true },
  callbacks = {
    on_global_turn = function(who, eff)
      -- Drain mana every turn [cf. src/dungeon.c lines 2896-2912]
      local dm = eff.power
      if rng.percent(30) then dm = dm + eff.power end
      who:incMana(-dm)
      if who.mana <= 0 and who.restStop then
	who:runStop('mana completely drained')
	who:restStop('mana completely drained')
      end
    end,
  },
  describe = function(who, p)
    local fmt = [[Your mana is being drained at a rate of %d-%d per normal-speed turn.]]
    return fmt:format(p.power, 2*p.power):toTString()
  end,
}

newEffect {
  name = 'DRAIN_LIFE',
  desc = 'Drain Life',
  display = function(eff) return 'Dr.Life' end,
  type = 'magical',
  status = 'detrimental',
  decrease = 0, -- Removed when we take off the equipment with the flag.
  no_dur = true, -- Don't show duration in the status display.
  equip_flags = { 'drain_hp', power=true },
  callbacks = {
    on_global_turn = function(who, eff)
      -- Drain hp every turn [cf. src/dungeon.c lines 2942-2959]
      local dl = eff.power + rng.range(0, math.floor(who.max_life/100) - 1)
      -- We do this manually to bypass :takeHit().
      who.life = math.max(0, who.life - dl)
      if who.mana <= 0 and who.restStop then
	who:runStop('life drained to zero')
	who:restStop('life drained to zero')
      end
    end,
  },
  describe = function(who, p)
    local fmt = [[Your life is being drained at a rate of %d-%dhp per normal-speed turn.  This will not drain your life below zero.]]
    return fmt:format(p.power, p.power + math.floor(who.max_life/100) - 1):toTString()
  end,
}

-- A copy of DRAIN_EXP with a different description for use by Maiar.
newEffect {
  name = 'MAIA_DRAIN_EXP',
  desc = 'Faded Memory',
  display = function(eff) return 'Dr.Exp' end,
  type = 'racial',
  status = 'detrimental',
  decrease = 0, -- Removed when we take off the equipment with the flag.
  no_dur = true, -- Don't show duration in the status display.
  -- This isn't an "equipment" flag per se, but it works with the new
  -- equipment flag infrastructure.
  equip_flags = { 'flags', 'MAIA_DRAIN_EXP' },
  callbacks = {
    on_global_turn = function(who, eff)
      -- 10% chance to drain 1exp every global turn
      -- [cf. src/dungeon.c lines 2962-2970]
      if who.exp and who.exp > 0 and rng.percent(10) then
	who:gainExp(-1)
	who.max_exp = math.max(who.max_exp - 1, who.exp)
      end
    end,
  },
  describe = function(who, p)
    return tstring {
      [[As one of the ancient race of Maiar, you sometimes have difficulty retaining your experiences.  Every normal-speed turn, you have a 10% chance of non-restorably losing 1 exp.]]
    }
  end,
}

newEffect {
  name = 'SPECTRE_IN_WALL',
  desc = 'In wall',
  display = function(eff) return 'InWall' end,
  src_desc = 'phasing through walls',
  type = 'physical',
  status = 'detrimental',
  cancel_on_level_change = true,
  decrease = 0, -- Lasts until we leave the grid
  no_dur = true, -- Don't show duration in the status display.
  terrain_effect = true, -- cf. Player:describeFloor()
  apply = function(who, p)
    who.no_life_regen = true
  end,
  on_timeout = function(who, p, def)
    local dam = 1 + math.floor((who.level or 0)/5)
    dam = math.min(dam, (who.life or 0) - 1)
    if dam > 0 then
      local src = p.src and util.effect_source(def, p.src) or util.effect_source(def, nil, true)
      who:takeHit(dam, src, 'phasing through solid matter')
    end
  end,
  on_gain = function(who, p)
    return 'This is going to hurt...', '+InWall'
  end,
  on_lose = function(who, p)
    return 'What a relief to be back out in the open.', '-InWall'
  end,
  describe = function(who, p)
    local dam = 1 + math.floor((who.level or 0)/5)
    local fmt = [[As a Spectre, you can force your way through solid walls, but it hurts.  You take %d damage per turn (increases with level) while you are in otherwise impassible terrain.  Note that this effect will never reduce you below 1hp.]]
    return tstring { fmt:format(dam) }
  end,
}

newEffect {
  name = 'ANTIMAGIC_FIELD',
  desc = 'Antimagic Field',
  display = function(eff) return 'Antimagic' end,
  type = 'antimagic',
  status = 'other',
  activate = function(who, p, def)
    print('[GRMBL] activate()')
    if config.settings.tome2.visible_antimagic_field then
      def.add_map_effect(who, p)
    end
  end,
  deactivate = function(who, p, def)
    -- We don't bother conditioning on the user pref here, because we've
    -- arranged for removing the effect if it's not present to be a no-op.
    def.remove_map_effect(who, p)
  end,
  decrease = 0, -- Removed when we take off the equipment with the flag.
  no_dur = true, -- Don't show duration in the status display.
  equip_flags = { 'antimagic', power=true },
  add_map_effect = function(who, p)
    print('[GRMBL] add_map_effect()')
    local MapEffect = require 'engine.MapEffect'
    local function cb(e)
      e.x = e.src.x
      e.y = e.src.y
      return true
    end
    local ov = MapEffect.new{ back_color={r=32,g=0,b=0} }
    p.map_eff = game.level.map:addEffect(who, who.x, who.y, 1, nil, nil,
					 who.antimagic_rad, 5, nil, ov, cb)
    p.map_eff.name = 'antimagic field'
    p.map_eff.decrease = 0
    p.map_eff.antimagic_field = true	-- to find and remove effect later
  end,
  remove_map_effect = function(who, p)
    local map, idx = game.level.map, nil
    -- Find our antimagic field effect.
    for i, e in ipairs(map.effects) do
      if e.antimagic_field then idx = i end
    end
    -- If we found it, remove it.
    if idx then
      -- Duplicated from Map:processEffects().
      local e = table.remove(map.effects, idx)
      if e.particles then
	for j, ps in ipairs(e.particles) do
	  map:removeParticleEmitter(ps)
	end
      end
      if e.overlay then
	map.z_effects[e.overlay.zdepth][e] = nil
      end
    end
    p.map_eff = nil
  end,
  callbacks = {
    on_leave_level = function(who, p, _, _, _, _)
      local def = game.player:getEffectFromId(who.EFF_ANTIMAGIC_FIELD)
      def.remove_map_effect(who, p)
    end,
    on_enter_level = function(who, p, _, _, _, _, _)
      if config.settings.tome2.visible_antimagic_field then
	local def = game.player:getEffectFromId(who.EFF_ANTIMAGIC_FIELD)
	def.add_map_effect(who, p)
      end
    end,
  },
  describe = function(who, p)
    local fmt = [[You are generating an antimagic field of radius %d.  While this field is active, you cannot use wands, staffs or rods, cast spells, or use powers conferred by the Mindcraft, Mimicry, Possession, Necromancy, Summoning or Symbiosis skills.  Other spellcasters attempting to cast spells within the field will fail with probability %d%%.  Probability and radius increase with your Antimagic skill level and when wielding antimagic equipment.]]
    return tstring{ fmt:format(who.antimagic_rad, who.antimagic) }
  end,
}

for _, stat in ipairs(Stats.stats) do
  local tag = 'LOSE_'..stat
  newEffect {
    name = tag,
    desc = 'Reduced '..stat,
    display = function(eff) return 'Lose '..stat end,
    type = 'physical',
    status = 'detrimental',
    on_gain = function(self, p)
      local msg, fly = nil, nil
      if not p.quiet then
	msg = 'You feel very '..Stats.stat_change_desc[stat][2]..'.'
	fly = '-'..stat
      end
      return msg, fly
    end,
    on_lose = function(self, p)
      local msg = 'You feel less '..Stats.stat_change_desc[stat][2]..'.'
      return msg, '+'..stat
    end,
    on_merge = util.merge_effect_duration,
    combine_if_merge = { 'penalty' },
    -- We don't need an apply() method here; this penalty is applied in
    -- ActorStats:getStat().
    describe = function(who, p)
      return tstring {
	('Your %s is temporarily reduced by %g.'):format(Stats.stat_names[stat], p.penalty/10)
      }
    end,
  }
end
