-- $Id: Store.lua 3549 2019-01-16 01:21:04Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Store = require 'engine.Store'
local Dialog = require 'engine.ui.Dialog'
local Object = require 'mod.class.Object'
local ShowStore = require 'mod.dialogs.ShowStore'
local GetQuantity = require "engine.dialogs.GetQuantity"

module(..., package.seeall, class.inherit(Store))

_M.stores_def = {}
_M.stock_levels = {
  -- NOTE:  armor/light doesn't collide with equipment/light because we
  -- check type before subtype.
  oil = 'food',
  light = 'food',
  food = {
    { cost=5, dice=3, sides=5 },
    { cost=20, dice=3, sides=5 },
  },
  scroll = 'potion',
  potion = {
    { cost=60, dice=3, sides=5 },
    { cost=240, dice=1, sides=5 },
  },
  spellbook = {
    { cost=50, dice=2, sides=3 },
    { cost=500, dice=1, sides=3 },
  },
  weapon = 'armor',
  shooter = 'armor',
  armor = {
    { cost=10, dice=3, sides=5 },
    { cost=100, dice=3, sides=5 },
  },
  spike = 'ammo',
  ammo = {
    { cost=5, dice=5, sides=5 },
    { cost=50, dice=5, sides=5 },
    { cost=500, dice=5, sides=5 },
  },
  staff = 'wand',
  wand = {
    { cost=1600, dice=1, sides=5 },
    { cost=3200, dice=1, sides=3 },
  },
}

function _M:loadStores(f)
  self.stores_def = self:loadList(f)
end

function _M:init(t, no_default)
  t.store.min_fill = math.min(t.store.min_fill or 6, t.max_stock or 24)
  t.store.max_fill = math.min(t.store.max_fill or 18, t.max_stock or 24)
  t.store.post_filter = t.store.post_filter or self.stockPostFilter
  t.store.purse = t.store.purse or 2500
  -- 1000 player turns at normal speed.
  t.store.restock_after = t.restock_after or 10000
  t.store.turnover = t.turnover or 9
  Store.init(self, t, no_default)
  -- There's no obvious way to override the 't.body = {INVEN=10000}' in
  -- Store.init(), so we patch that up here.
  self:getInven('INVEN').max = t.max_stock or 24
  -- Can't sell to stores in dungeon towns.
  if t.store.dungeon_town then self.allow_sell = false end
end

function _M.stockPostFilter(o)
  -- Vaguely Annoying Hack(TM):  require() ourself so that we can actually
  -- access our own @#$%ing fields...
  local Store = require 'mod.class.Store'
  -- Stock piles of some items.
  local t = Store.stock_levels[o.type] or Store.stock_levels[o.subtype]
  while type(t) == 'string' do t = Store.stock_levels[t] end
  if type(t) == 'table' then
    -- TODO Don't multi-stock ego items.
    local size = 1
    for _, tt in ipairs(t) do
      if o.cost <= tt.cost then size = size + rng.dice(tt.dice, tt.sides) end
    end
    o:setNumber(size)
  end

  return true
end

function _M:getObjectPrice(who, o, qty, buying)
  local owner_cap = self.owner and self.owner.max_cost or nil
  local price = o:getPrice(qty, {unit_max=owner_cap})
  if price < 0 then return 0 end

  local adj = who:getStatRelated('cha_price')
  if self.owner then
    adj = adj + self.owner.inflate + self.owner:getRaceAdjust(who)
  end
  if buying then
    adj = math.max(100, 100 + (adj - 300))
    -- Black market prices suck.
    if self.black_market then price = price*2 end
  else
    -- T2 describes this as a "Mega Hack^3"
    if o.type == 'ammo' then price = math.floor(price/5) end
    adj = math.min(100, 100 + (300 - adj))
    -- Black market prices suck.
    if self.black_market then price = math.floor(price/2) end
  end

  return math.max(1, math.floor((price*adj + 50)/100))
end

function _M:tryBuy(who, o, item, qty)
  local total_price = self:getObjectPrice(who, o, qty, true)
  if who.gold >= total_price then
    return qty, total_price
  else
    Dialog:simplePopup('Not enough gold', 'Liar!  You have not the gold!')
    return nil, nil
  end
end

function _M:trySell(who, o, item, qty)
  local total_price = self:getObjectPrice(who, o, qty, false)
  -- I thought I remembered the stores having a purse that they could run
  -- out of, but I can't find it in the code.
  if total_price <= 0 then
    Dialog:simplePopup('No sale', 'I will not buy that!')
    return nil, nil
  end

  -- Do we have room in the store?
  local inven = self:getInven('INVEN')
  local can_add = #inven < inven.max
  if not can_add and o and o:stackable() then
    -- Maybe we can stack it with something?
    local stack_limit = self:invenStackLimit('INVEN')
    for i, io in ipairs(inven) do
      if o:canStack(io) and io:getNumber() + qty <= stack_limit then
	can_add = true
	break
      end
    end
  end
  if not can_add then
    Dialog:simplePopup('Store full', 'I have no room for that in my store.')
    return nil, nil
  end

  return qty, total_price
end

function _M:onBuy(who, o, item, qty, before)
  if before then return end
  local total_price = self:getObjectPrice(who, o, qty, true)
  if who.gold >= total_price then
    who:adjGold(-total_price)
    local o_name = o:getName{ force_qty=qty }
    game.log('%s bought %s for %s gold.', who.name, o_name, total_price)
  end
end

function _M:onSell(who, o, item, qty, total_price, before)
  if not before then
    who:adjGold(total_price)
    local o_name = o:getName{ force_qty=qty, force_id=true }
    game.log('%s sold %s for %s gold.', who.name, o_name, tostring(total_price))
  end
end

-- We replace the parent transfer() to support dropping the "bought" object
-- if it doesn't fit into the player's inventory, to identify sold items,
-- and to insert/remove found{} info for transferred objects.  [We can't do
-- the latter two in onBuy()/onSell() because we need a ref to the object
-- in dest for that.]
function _M:transfer(src, dest, item, nb, buying)
  local src_inven, dest_inven = src:getInven("INVEN"), dest:getInven("INVEN")
  -- Annoying Hack(TM):  Borrow the 'stocking' flag to prevent alchemy
  -- recipe learning on anything we're about to sell.
  _M.stocking = true
  -- Selling one item off a stack should ID the stack.  This is mainly for
  -- ammo, to get the (+x,+y).
  if not buying and src_inven[item] and not src_inven[item]:isIDd() then
    src_inven[item]:identify(true)
  end
  for i = 1, nb do
    local o = src:removeObject(src_inven, item)
    if buying then
      -- Set found{} info
      o:setFoundHere{ type='store_buy', store_name=self.name }
      dest:addOrDropObject(dest_inven, o, nil, nil, {vocal_add=true, vocal_drop=true})
    else
      -- Clear found{} info
      o.found = nil
      o:identify('full')
      dest:addObject(dest_inven, o)
    end
  end
  -- Clean this up.
  _M.stocking = nil
  src:sortInven(src_inven)
  dest:sortInven(dest_inven)
end

function _M:doBuy(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  local total_price
  qty, total_price = self:tryBuy(who, o, item, qty)
  if qty then
    local o_name = o:getName{ force_qty=qty }
    local function callback(ok)
      if (ok) then
	self:onBuy(who, o, item, qty, true)
	self:transfer(self, who, item, qty, true)
	self:onBuy(who, o, item, qty, false)
	if store_dialog then store_dialog:updateStore() end
      end
    end
    Dialog:yesnoPopup('Buy', ('Buy %s for %d gold?'):format(o_name, total_price), callback, 'Buy', 'Cancel')
  end
end

function _M:doSell(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  local total_price
  qty, total_price = self:trySell(who, o, item, qty)
  if qty then
    local o_name = o:getName{ force_qty=qty }
    local function callback(ok)
      if (ok) then
	self:onSell(who, o, item, qty, total_price, true)
	self:transfer(who, self, item, qty, false)
	self:onSell(who, o, item, qty, total_price, false)
	if store_dialog then store_dialog:updateStore() end
      end
    end
    Dialog:yesnoPopup('Sell', ('Sell %s for %d gold?'):format(o_name, total_price), callback, 'Sell', 'Cancel')
  end
end

function _M:descObjectPrice(who, what, o)
  return ('%d'):format(self:getObjectPrice(who, o, 1, what == 'buy'))
end

-- Loosely based on store_create() in src/store.c
function _M:allowStockObject(o)
  local price = o.getPrice and o:getPrice(true, {force_id=true})
  -- Don't allow cursed stuff.
  if o.cursed then return false end
  -- No chests.
  if o.type == 'misc' and o.subtype == 'chest' then return false end
  if self.store and self.store.black_market then
    -- In black market, reject "cheap" stuff...
    if price and price < 10 then return false end
    -- ...and "crappy" stuff.
    if self:isBlackMarketReject(o) then return false end
  end
  -- Don't allow "bad" (ie. cost < 0) stuff.
  return price and price > 0
end

--- Fill the store with goods
-- Overloaded to handle T2's idiosyncratic filter selection, and our
-- Object.resolve_flags{} juggling.
-- @param level the level to generate for (instance of type engine.Level)
-- @param zone the zone to generate for
function _M:loadup(level, zone, force_nb)
  local s = self.store
  if not s then error("Store without a store field") end
  if not self:canRestock() then return end
  local inven = self:getInven("INVEN")

  -- Annoying Hack(TM):  Let downstream code know we're stocking stores, so
  -- that, for instance, ID'ing stock won't trigger alchemy recipe learning.
  _M.stocking = true

  local i = 1
  local rngfill = force_nb or (rng.range(s.min_fill, s.max_fill) - #inven)
  while i <= rngfill do
    local filter = self:random_filter(s.prob_filters)
    if not filter then break end
    filter = table.clone(filter)
    self:setup_resolve_flags(s)
    local e
    if not filter.defined then
      e = zone:makeEntity(level, "object", filter, nil, true)
    else
      e = zone:makeEntityByName(level, "object", filter.defined)
    end
    Object.resolve_flags = {}
    if e and not e.not_in_stores and self:allowStockObject(e) then
      if s.post_filter and not s.post_filter(e) then
      else
	if filter.id then e:identify(filter.id) end
	self:addObject(inven, e)
	zone:addEntity(level, e, "object")
	self:check("stocked_object", e)
	print("[STORE] stocking up: ", e.name)
	i = i + 1
      end
    elseif not e then
      -- If resolving failed, we probably need to clear these manually.
      Object.resolve_flags = {}
      i = i + 1
    end
  end

  for i = 1, #(s.fixed or {}) do
    local filter = table.clone(s.fixed[i])
    self:setup_resolve_flags(s)
    local e
    if not filter.defined then
      e = zone:makeEntity(level, "object", filter, nil, true)
    else
      e = zone:makeEntityByName(level, "object", filter.defined)
    end
    if e and not e.not_in_stores and self:allowStockObject(e) then
      if filter.id then e:identify(filter.id) end
      self:addObject(inven, e)
      zone:addEntity(level, e, "object")
      self:check("stocked_object", e)
      print("[STORE] stocking up: ", e.name)
    elseif not e then
      -- As above.
      Object.resolve_flags = {}
    end
  end

  -- And we're done with this.
  _M.stocking = nil

  self:sortInven(inven)
  self.last_filled = game.turn
  return true
end

-- Selects a probability-weighted filter out of the provided filters{}
-- table.
function _M:random_filter(filters)
  if not filters then
    error('No prob_filters field in store loadup()')
    return nil
  end
  local ret = nil
  for _ = 1, 50 do
    local f = rng.table(filters)
    if not f.prob or rng.percent(f.prob) then
      ret = f
      break
    end
  end
  if not ret then
    -- If we didn't get one after fifty tries, force the issue and randomly
    -- pick one of the 100%-probability ones.
    local t100 = {}
    for _, f in ipairs(filters) do
      if not f.prob or f.prob >= 100 then table.insert(t100, f) end
    end
    ret = rng.table(t100)
  end
  return ret
end

-- Set up Object.resolve_flags{} based on the provided store.
function _M:setup_resolve_flags(s)
  local RF = Object.resolve_flags
  local lev = s.dungeon_town and 0 or rng.range(1, 5)
  if s.use_dungeon_level then
    lev = lev + game.zone:level_adjust_level(game.level, 'object')
  end
  if s.level_adjust then
    lev = lev + rng.range(s.level_adjust[1], s.level_adjust[2])
  end
  if s.black_market then lev = lev + game.player.level end
  RF.force_level = lev
  RF.store_pop = true
  if s.min_obj_level then RF.store_min_level = math.floor(lev/2) end
  if s.resolve_flags then table.merge(RF, s.resolve_flags, true) end
end

-- Vaguely Annoying Hack(TM) to allow the store.actor_filter() method to
-- access the stock_base{} field of the store itself, which doesn't get
-- passed to the method.
-- Static
function _M:actor_filter_hack(tag, o)
  local e = self.stores_def[tag]
  local base = e and e.store and e.store.stock_base
  if not base then return false end
  if type(base) == 'table' and base[o.type] then
    local x = base[o.type]
    if type(x) == 'boolean' then return true end
    if type(x) == 'string' then return x == o.subtype end
    if type(x) == 'function' then return x(o) end
    if type(x) == 'table' and x[o.subtype] then
      local xx = base[o.type][o.subtype]
      if type(xx) == 'boolean' then return true end
      if type(xx) == 'function' then return xx(o) end
    end
  end
  if type(base) == 'function' then return base(o) end
  return false
end

-- Loosely based on store_maint() in src/store.c [q.v.]
function _M:cycleStock()
  local s = self.store
  local inven = self:getInven('INVEN')
  local n_cycle = math.floor((game.turn - self.last_filled)/s.restock_after)
  n_cycle = math.min(n_cycle, 10)
  if #inven == 0 then n_cycle = 10 end
  if n_cycle <= 0 then return end
  self.last_filled = game.turn

  for _ = 1, n_cycle do
    if s.black_market then
      -- Ditch anything in the black market that they wouldn't sell.
      for item = #inven, 1, -1 do
	if inven[item] and self:isBlackMarketReject(inven[item]) then
	  self:removeObject(inven, item, true)
	end
      end
    end
    local tgt_stock = util.bound(#inven - rng.range(1, s.turnover), s.min_fill, s.max_fill)
    while #inven > tgt_stock do
      -- cf. store_delete() in src/store.c
      local n, item = true, rng.range(1, #inven)
      local o = inven[item]
      if o:getNumber() > 1 then
	if rng.percent(50) then n = math.ceil(o:getNumber()/2) end
	if rng.percent(50) then n = 1 end
      end
      -- We don't call :removed() on the objects we remove, as that would
      -- put sold artifacts back into circulation.
      self:removeObject(inven, item, n)
    end
    tgt_stock = util.bound(#inven + rng.range(1, s.turnover), s.min_fill, s.max_fill)
    if tgt_stock < #inven then
      self:loadup(game.level, game.zone, #inven - tgt_stock)
    end
  end
end

-- Returns true if the provided object should not be stocked in a store
-- flagged 'black_market'.
-- [cf. black_market_crap() in src/store.c]
function _M:isBlackMarketReject(o)
  -- Don't reject ego items.
  if o.ego_names then return false end
  -- Don't reject anything with any positive enchantment.
  for _, f in ipairs({'to_a', 'to_h', 'to_d'}) do
    if o[f] and o[f] > 0 then return false end
  end

  -- Reject anything of the same type as something being sold in any store
  -- in town. [Oddly, AFAICT in the T2 code this includes the same store
  -- we're in the middle of populating...]
  for _, store in ipairs(game:getStores('store')) do
    local inven = store:getInven('INVEN')
    for _, oo in ipairs(inven) do
      -- The T2 code uses the 'k_idx' field here, which we don't retain;
      -- 'define_as' should be sufficiently unique for our purposes.
      if oo.define_as and oo.define_as == o.define_as then return true end
    end
  end

  -- Looks like we're good.
  return false
end

-- Mostly the same as engine.Store.interact(), with a few tweaks.
function _M:interact(who, name)
  self:cycleStock()
  local d
  local store, inven = self:getInven("INVEN"), who:getInven("INVEN")
  local function action_cb(what, o, item)
    local o_name = o:getName()
    if what == "buy" then
      if o:getNumber() > 1 then
	local q = GetQuantity.new('Buying: '..o_name, 'Number to buy', o:getNumber(), o:getNumber(), function(qty) self:doBuy(who, o, item, qty, d) end, 1)
	game:registerDialog(q)
      else
	self:doBuy(who, o, item, 1, d)
      end
    else
      if o:getNumber() > 1 then
	local q
	q = GetQuantity.new('Selling: '..o_name, 'Number to sell', o:getNumber(), o:getNumber(), function(qty) self:doSell(who, o, item, qty, d) end, 1)
	game:registerDialog(q)
      else
	self:doSell(who, o, item, 1, d)
      end
    end
  end
  local function desc_cb(what, o)
    return self:descObject(who, what, o)
  end
  local function descprice_cb(what, o)
    return self:descObjectPrice(who, what, o)
  end
  d = ShowStore.new("Store: "..(name or self.name), self, store, who, inven,
		    self.store.store_filter, self.store.actor_filter,
		    action_cb, desc_cb, descprice_cb,
		    self.allow_sell, self.allow_buy,
		    function(item) end, self, who)
  game:registerDialog(d)
end

function _M:buyActionName()
  return 'Buy'
end

function _M:sortInven(inven)
  if not inven then inven = self.inven[self.INVEN_INVEN] end
  inven = self:getInven(inven)
  if not inven then return end
  obj_util.sort_inven(inven)
  self.changed = true
end
