-- $Id: ActorLearn.lua 2835 2017-04-11 00:13:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'

module(..., package.seeall, class.make)

function _M:init(t)
end

function _M:rawData(use_ego)
  local raw = game.zone.npc_list[self.define_as] or {}
  local ego = (type(use_ego) == 'string' and use_ego) or (type(use_ego) == 'boolean' and use_ego and self.ego) or nil
  if ego then
    -- TODO Modify for ego
  end
  return raw
end

-- Static
function _M:learnedDataFor(tag, ego)
  local prepop = false
  world.mon_learn = world.mon_learn or { __wire_version=1 }
  local ret = world.mon_learn[tag]
  if not ret then
    world.mon_learn[self.define_as] = { flags={} }
    ret = world.mon_learn[self.define_as]
    prepop = true
  end
  if prepop then
    -- Prepopulate some obvious flags.
    local flags = {
      'UNIQUE', 'MALE', 'FEMALE', 'FRIEND', 'FRIENDS', 'ESCORT', 'ESCORTS',
    }
    local raw = game.zone.npc_list[tag] or {}
    for _, f in ipairs(flags) do ret.flags[f] = raw.flags[f] end
  end
  if ego then
    -- TODO Modify for ego
  end
  return ret
end

function _M:learnedData(raw)
  local ego = self.ego
  if raw then ego = nil end
  return self:learnedDataFor(self.define_as, ego)
end

function _M:learnFlags(force, grp, ...)
  -- Mildly Convoluted Hack(TM):  If the first parameter is not a boolean
  -- 'force' flag, pretend we received force=false.
  if _G.type(force) ~= 'boolean' then
    self:learnFlags(false, force, grp, ...)
    return
  end

  if not self.player and (force or (game.player:hasLOS(self.x, self.y) and game.player:canSee(self)) or (self.symbiosis and self.symbiosis.host.player)) then
    self:learnFlagsFor(self.define_as, grp, ...)
  end
end

-- Static
function _M:learnFlagsFor(tag, grp, ...)
  -- Occasionally happens when loading a save file.
  if not game.zone or not game.zone.npc_list then return end

  local args = {...}
  local raw = game.zone.npc_list[tag] or {}
  local learned = self:learnedData(raw)
  if grp == 'attacks' then
    if not raw.attacks then return end
    if not learned.attacks then
      learned.attacks = table.clone(raw.attacks, true)
      for _, atk in ipairs(learned.attacks) do atk.hit_count = 0 end
    end
    local idx, valid = ...
    if not learned.attacks[idx] then return end
    if valid or learned.attacks[idx].hit_count > 10 then
      learned.attacks[idx].hit_count = math.min(learned.attacks[idx].hit_count + 1, 255)
    end
  elseif grp == 'sleep' then
    learned.sleep = learned.sleep or { ignore=0, wake=0 }
    local which = args[1]
    -- "Woke-up"/"didn't-wake-up" counts are capped at MAX_UCHAR=255,
    -- as per process_monster() in src/melee2.c
    learned.sleep[which] = math.min(learned.sleep[which] + 1, 255)
  elseif grp == 'drops' then
    -- Remember max number of gold/items we've seen the subject drop.
    local n_item, n_gold = ...
    learned.drop_desc = learned.drop_desc or { items=0, gold=0 }
    learned.drop_desc.items = math.max(learned.drop_desc.items, n_item)
    learned.drop_desc.gold = math.max(learned.drop_desc.gold, n_gold)
  else
    for _, v in ipairs(args) do
      learned[grp] = learned[grp] or {}
      learned[grp][v] = raw[grp] and raw[grp][v]
    end
    if grp == 'talents' then
      -- Remember how many times we've seen the subject use talents.
      learned.seen = learned.seen or {}
      learned.seen.talent = (learned.seen.talent or 0) + #args
    end
  end
end

-- Static
function _M:learnAllFlagsFor(tag)
  local learned = self:learnedDataFor(tag)
  local raw = game.zone.npc_list[tag] or {}
  local flag_grps = {
    'flags', 'resist_dam', 'vuln_dam', 'immune_dam', 'talents',
  }
  for _, grp in ipairs(flag_grps) do
    learned[grp] = learned[grp] or {}
    for f, v in pairs(raw[grp]) do learned[grp][f] = v end
  end
  learned.attacks = table.clone(raw.attacks, true)
  for _, atk in ipairs(learned.attacks) do atk.hit_count = 255 end
  learned.sleep = { ignore=255, wake=255 }
  learned.drop_desc = table.clone(raw.drop_desc, true)
  learned.seen = { talent=255 }
end

function _M:recordKill()
  if not self.define_as then return end

  local pks = table.getTable(game.player, 'kills')
  pks.all = (pks.all or 0) + 1
  pks[self.define_as] = (pks[self.define_as] or 0) + 1

  local wks = table.getTable(world, 'kills')
  wks.all = (wks.all or 0) + 1
  if not wks[self.define_as] then
    -- First kill earns us some learned data
    local data = self:learnedData()
    data.flags = data.flags or {}
    local flags = {
      'ORC', 'TROLL', 'GIANT', 'DRAGON', 'DEMON', 'UNDEAD', 'EVIL', 'GOOD',
      'ANIMAL', 'THUNDERLORD', 'SPIDER', 'FORCE_DEPTH', 'FORCE_MAX_DEPTH',
    }
    for _, f in ipairs(flags) do
      data.flags[f] = self.flags and self.flags[f]
    end
  end
  wks[self.define_as] = (wks[self.define_as] or 0) + 1
end

function _M:killCount()
  if not self.define_as then return 0, 0 end
  local p_kills = game.player.kills and game.player.kills[self.define_as] or 0
  local w_kills = world.kills and world.kills[self.define_as] or 0
  return p_kills, w_kills
end

function _M:recordPlayerKill()
  if not self.define_as then return end

  world.killed_by = world.killed_by or { __wire_version=1 }
  world.killed_by[self.define_as] = (world.killed_by[self.define_as] or 0) + 1
end

function _M:playerKillCount()
  if not self.define_as then return 0 end
  return world.killed_by and world.killed_by[self.define_as] or 0
end
