-- $Id: PlayerSkills.lua 2864 2017-05-08 01:36:27Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'

module(..., package.seeall, class.make)

_M.skill_step = 1000
_M.skill_max = 50*_M.skill_step
_M.skills_def = {}
_M.skills_toplevel = {}
_M.abilities_def = {}
_M.abilities_order = {}

function _M:loadDefinition(file, env)
  env = env or {
    Skills = self,
    ActorTalents = require 'engine.interface.ActorTalents',
    DamageType = require 'engine.DamageType',
    newSkill = function(t) self:newSkill(t) end,
    newAbility = function(t) self:newAbility(t) end,
    load = function(f) self:loadDefinition(f) end
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err
    then error(err)
    os.exit()
  end
  local ok, err = pcall(f)
  if not ok and err then error(err) end
end

function _M:newSkill(t)
  assert(t.name, 'no skill name')
  t.short_name = t.short_name or t.name:upper():gsub('[^a-zA-Z0-9_]', '_')
  self.skills_def[t.short_name] = t

  if t.parent then
    assert(self.skills_def[t.parent], 'undefined parent '..t.parent)
    local parent = self.skills_def[t.parent]
    parent.children = parent.children or {}
    table.insert(parent.children, t.short_name)
  else
    table.insert(self.skills_toplevel, t.short_name)
  end
end

function _M:newAbility(t)
  assert(t.name, 'no ability name')
  assert(t.cost, 'no ability cost')
  t.short_name = t.short_name or t.name:upper():gsub('[^a-zA-Z0-9_]', '_')
  self.abilities_def[t.short_name] = t
  table.insert(self.abilities_order, t.short_name)
end

function _M:getSkillDefinition(tag)
  return self.skills_def[tag]
end

function _M:init(t)
  self.skill_points = 0
  self.skills = {}
  self.abilities = {}
  self.bought_abilities = {}
  for tag, def in pairs(self.skills_def) do
    local t = {
      value = 0,
      mod = 0,
      name = def.name,
      desc = def.desc,
    }
    self.skills[tag] = t
  end
end

function _M:getSkillLevel(tag, scale, unrounded)
  local sk = self.skills[tag]
  if not sk then return 0 end
  local ret = (sk.value * (scale or 50)) / _M.skill_max
  return unrounded and ret or math.floor(ret)
end

function _M:setSkillValue(tag, value)
  local sk = self.skills[tag]
  if sk then
    local def = self.skills_def[tag]
    local old_val = sk.value
    sk.value = math.min(value, _M.skill_max)
    if sk.value >= _M.skill_max then
      world:gainAchievement(def.achievement, self)
    end
    if sk.value > old_val and def.on_increase then
      def:on_increase(self, old_val, sk.value)
    end
  end
end

function _M:adjSkillValue(tag, delta)
  local sk = self.skills[tag]
  if sk then self:setSkillValue(tag, sk.value + delta) end
end

function _M:preRecalcForSkills()
  self.target_talents = {}
  for _, def in pairs(self.skills_def) do
    if def.player_calc then def:player_calc(self, true) end
  end
  self.gloves_lower_mana = nil
end

function _M:recalcForSkills()
  for _, def in pairs(self.skills_def) do
    local tag = def.short_name
    local sk = self:getSkillLevel(tag)
    local sk_raw = self:getSkillLevel(tag, nil, true)

    if def.player_calc then def:player_calc(self) end
    if def.no_gloves and sk_raw > 0 then self.gloves_lower_mana = true end
    if def.talents then
      -- Elegant Hack(TM):  To avoid having talents hotkeyed in random
      -- order if we get more than one at once, the skill definitions list
      -- talents in numerical order and maps talent requirements by talent
      -- IDs.
      for _, tid in ipairs(def.talents) do
	local req = def.talents[tid]
	if type(req) == 'number' then req = { level = req } end
	if type(req) == 'boolean' then req = { any = req } end
	local add = true
	if req.level then add = add and sk >= req.level end
	if req.any then add = add and sk_raw > 0 end
	if req.ability then add = add and self.abilities[req.ability] end
	for otag, olev in pairs(req.skills or {}) do
	  add = add and self:getSkillLevel(otag) >= olev
	end
	if add then self:accumTargetTalent(tid) end
      end
    end
    self:applyLeveledFlagBundles(def.convey_at_level, sk)
  end
  for tag, _ in pairs(self.abilities) do
    local ab = _M.abilities_def[tag]
    if ab.talent then self:accumTargetTalent(ab.talent) end
    for f, _ in pairs(ab.convey_flags or {}) do self.flags[f] = true end
  end
end

function _M:accumTargetTalent(tid)
  if not self.target_talents[tid] then table.insert(self.target_talents, tid) end
  self.target_talents[tid] = true
end
