-- $Id: GameOptions.lua 3488 2018-10-27 22:55:41Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Tabs = require 'engine.ui.Tabs'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Textzone = require 'engine.ui.Textzone'
local Button = require 'engine.ui.Button'
local GetQuantity = require 'engine.dialogs.GetQuantity'

module(..., package.seeall, class.inherit(Dialog))

local default_desc = '{No description for this option.}'
local placeholder_desc = '\n\n#RED#This option is not yet implemented.  We Apologize for the Inconvenience.™#LAST#'
local restart_desc = '\n\n#YELLOW#You must restart the game for this option to take effect.'
local birth_desc = '#YELLOW##{italic}#This is a birth option; its value cannot be changed after the character is created.#{normal}##LAST#\n\n'

_M.opt_yes = string.toTString('#00FF00#Y#LAST#')
_M.opt_no = string.toTString('#FF0000#N#LAST#')
_M.opt_yn = { [true] = _M.opt_yes, [false] = _M.opt_no }

_M.options_def = {}
_M.option_groups = {}
_M.options_order = {}

function _M:loadDefinition(file, env)
  env = env or {
    GameOptions = self,
    newOption = function(t) self:newOption(t) end,
    newOptionGroup = function(t) self:newOptionGroup(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

function _M:newOption(t)
  assert(t.key, 'no option key')
  assert(t.name, 'no option name')
  assert(t.group, 'no option group')
  assert(self.option_groups[t.group], "unrecognized option group '"..t.group.."'")
  assert(not self.options_def[t.key], "option '"..t.key.."' already exists")
  assert(type(t.default) ~= 'nil', 'no option default')
  if type(t.default) == 'number' then
    assert(t.min and t.max, 'no option min/max value')
  elseif type(t.default) == 'string' then
    assert(t.values or t.dialog, 'need values{} or dialog() for string option')
    if t.values then
      local ok = false
      for _, val in ipairs(t.values) do ok = ok or val == t.default end
      assert(ok, 'default value not in values{}')
    end
  elseif type(t.default) ~= 'boolean' then
    local typ = type(t.default)
    assert(t.dialog, 'dialog() needed for '..typ..' option value')
    assert(t.display_val, 'display_val() needed for '..typ..' option value')
  end
  t.desc = t.desc or default_desc
  if t.placeholder then t.desc = t.desc..placeholder_desc end
  if t.require_restart then t.desc = t.desc..restart_desc end
  self.options_def[t.key] = t
  table.insert(self.options_order, t.key)
  table.insert(self.options_order[t.group], t.key)
end

function _M:newOptionGroup(t)
  assert(t.key, 'no option group key')
  assert(t.name, 'no option group name')
  assert(not self.option_groups[t.key], "option group '"..t.key.."' already exists")
  assert(not t.before or not t.after, "'before' and 'after' cannot both be set")

  local idx = #self.option_groups+1
  local ref = t.before or t.after
  if ref then
    for i, grp in ipairs(self.option_groups) do
      if grp == ref then
	idx = i + (t.after and 1 or 0)
	break
      end
    end
  end
  table.insert(self.option_groups, idx, t.key)
  self.option_groups[t.key] = t.name
  self.options_order[t.key] = {}
end

function _M:setupOptions()
  config.settings.tome2 = config.settings.tome2 or {}
  for key, info in pairs(self.options_def) do
    local val = config.settings.tome2[key]
    if type(val) == 'nil' then
      -- This both fills out an empty settings table and fills in default
      -- values for any new options added since the last time settings were
      -- saved.
      if type(info.default) == 'table' then
	config.settings.tome2[key] = table.clone(info.default, true)
      else
	config.settings.tome2[key] = info.default
      end
    elseif info.legacy_convert and info.legacy_convert[val] then
      -- Slight Hack(TM) to convert legacy setting values
      if type(info.legacy_convert[val]) == 'table' then
	config.settings.tome2[key] = table.clone(info.legacy_convert[val], true)
      else
	config.settings.tome2[key] = info.legacy_convert[val]
      end
    end
  end
end

local function buildOptionSave(t, ind)
  ind = ind or ''
  local ret = '{\n'
  for k, v in pairs(t) do
    if type(v) == 'boolean' then
      ret = ret .. ind ..  ('  %s = %s,\n'):format(k, v and 'true' or 'false')
    elseif type(v) == 'number' then
      ret = ret .. ind ..  ('  %s = %d,\n'):format(k, v)
    elseif type(v) == 'string' then
      ret = ret .. ind ..  ('  %s = %q,\n'):format(k, v)
    elseif type(v) == 'table' then
      local vv = buildOptionSave(v, ind..'  ')
      ret = ret .. ind ..  ('  %s = %s,\n'):format(k, vv)
    else
      print('[OPTIONS] unrecognized option type', k, v)
    end
  end
  return ret .. ind .. '}'
end

function _M:saveOptions()
  local save = 'tome2 = ' .. buildOptionSave(config.settings.tome2) .. '\n';
  game:saveSettings('tome2', save)
end

-- Save birth option settings into the game state, so that changing them
-- for another character doesn't affect this one.  This is called from the
-- Birther dialog, early enough that they should be in place for everyone
-- who needs them.
function _M:keepBirthOptions()
  local birth = {}
  for key, def in pairs(self.options_def) do
    if def.group == 'birth' then
      birth[key] = config.settings.tome2[key]
    end
  end
  game.state.birth = birth
end

function _M:init()
  Dialog.init(self, 'Game Options', 600, 400)
  local orig_opt = table.clone(config.settings.tome2, true)

  self.c_ok = Button.new {
    text = 'Apply',
    fct = function()
      game:unregisterDialog(self)
      self:saveOptions()
      game.player.changed = true
    end
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function()
      config.settings.tome2 = orig_opt
      game:unregisterDialog(self)
      game.player.changed = true
    end
  }

  self.c_ok:generate()
  local main_h = self.ih - self.c_ok.h - 5
  self.c_desc = Textzone.new {
    width = math.floor(self.iw/2 - 10),
    height = main_h,
    scrollbar = true,
    no_color_bleed = true,
    text = ''
  }
  self.c_list = self:createOptionsList(math.floor(self.iw/2 - 10), main_h, self.c_desc)

  local tabs = {}
  for _, tag in ipairs(self.option_groups) do
    table.insert(tabs, { title=self.option_groups[tag], kind=tag })
  end
  self.c_tabs = Tabs.new {
    width = self.iw - 5, 
    tabs = tabs,
    on_change = function(tag) self:switchTo(tag) end,
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_tabs },
    { left=0, top=self.c_tabs.h, ui=self.c_list },
    { right=0, top=self.c_tabs.h, ui=self.c_desc },
    { right=0, top=self.c_desc, ui=self.c_cancel },
    { right=self.c_cancel, top=self.c_desc, ui=self.c_ok },
  }
  self:setFocus(self.c_list)
  self:setupUI()

  self.key:addCommands {
    _TAB = function() self:tabTabs() end,
  }
  self.key:addBinds { EXIT = self.c_cancel.fct }
end

function _M:on_register()
  game:onTickEnd(function() self.key:unicodeInput(true) end)
end

function _M:switchTo(grp)
  self.c_list:populateForGroup(grp)
  local item = self.c_list.list[self.c_list.sel]
  if not item then
    self.c_list.sel = 1
    item = self.c_list.list[1]
  end
  self.c_desc.text = self:computeOptionDesc(item.key, false)
  self.c_desc:generate()
end

function _M:tabTabs()
  local sel = nil
  for i, ui in ipairs(self.c_tabs.uis) do
    if ui.ui.selected then sel = i end
  end
  if sel then
    sel = util.boundWrap(sel+1, 1, #self.c_tabs.tabs)
    self.c_tabs:select(self.c_tabs.tabs[sel].kind)
  end
end

-- [static] Create a TreeList object that manipulates game options and
-- puts text descriptions in the provided Textzone.
function _M:createOptionsList(w, h, text, set_cb, birther)
  local c_list
  c_list = TweakedTreeList.new {
    simple_list = true,
    width = w,
    height = h,
    scrollbar = true,
    columns = {
      { name='Option', width=80, display_prop='name' },
      { name=' ', width=20, display_prop='display_val' },
    },
    tree = {},
    fct = function(item, _, _)
      self:alterOption(item, c_list, set_cb, birther)
    end,
    select = function(item, sel)
      text.text = self:computeOptionDesc(item.key, birther)
      text:generate()
    end
  }
  if birther then
    local opts = {}
    for _, key in ipairs(self.options_order) do
      local def = self.options_def[key]
      if def.group == 'birth' or def.show_in_birther then
	table.insert(opts, key)
      end
    end
    self:generateList(opts, c_list)
  else
    c_list.populateForGroup = function(list_self, grp)
      list_self.read_only = grp == 'birth'
      self:generateList(self.options_order[grp], list_self)
    end
  end
  return c_list
end

-- [static] Called from the TreeList created by createOptionsList() to
-- populate the list with the provided options.
function _M:generateList(keys, c_list)
  local list = {}

  for _, key in ipairs(keys) do
    local def = self.options_def[key]
    local v = config.settings.tome2[key]
    local dv = v
    if def.display_val then
      dv = def:display_val(v)
    elseif type(v) == 'boolean' then
      dv = self.opt_yn[v]
    elseif def.short_values then
      dv = def.short_values[v]
    end

    list[#list+1] = { key=key, name=def.name, value=v, display_val=dv }
  end

  c_list.tree = list
  c_list:generate()
end

-- [static] Called from the TreeList created by createOptionsList() to
-- actually set option values.
function _M:alterOption(item, c_list, set_cb, birther)
  if not item or c_list.read_only then return end
  local def = self.options_def[item.key]
  if def.dialog then
    -- Notes:  An option definition can have a dialog() method to input
    -- complex methods.  It will receive two arguments, the option
    -- definition table and the current option value, and should return the
    -- created Dialog.  The created dialog must have methods :get_value()
    -- and :get_display_val(); these should return nil if the user canceled
    -- out of the dialog, otherwise the new value and a short display
    -- string for use in the main GameOptions dialog.
    local d = def:dialog(config.settings.tome2[item.key])
    d.unload = function(dlg_self)
      local val = dlg_self:get_value()
      if type(val) ~= 'nil' then
	item.value = val
	item.display_val = dlg_self:get_display_val()
	config.settings.tome2[item.key] = val
	c_list:drawItem(item)
	if def.on_change then def.on_change(item.value, birther) end
	if set_cb then set_cb(item.key) end
      end
    end
    game:registerDialog(d)
  elseif type(item.value) == 'boolean' then
    item.value = not item.value
    item.display_val = self.opt_yn[item.value]
    config.settings.tome2[item.key] = item.value
    c_list:drawItem(item)
    if def.on_change then def.on_change(item.value, birther) end
    if set_cb then set_cb(item.key) end
  elseif type(item.value) == 'number' then
    local min = self.options_def[item.key].min
    local max = self.options_def[item.key].max
    local callback = function(qty)
      qty = util.bound(qty, min, max)
      item.value = qty
      item.display_val = qty
      config.settings.tome2[item.key] = qty
      c_list:drawItem(item)
      if def.on_change then def.on_change(item.value, birther) end
      if set_cb then set_cb(item.key) end
    end
    local d = GetQuantity.new(item.name, item.name, item.value, max, callback, min)
    -- Bizarre Hack(TM):  For no reason I can fathom, this needs to be done
    -- here but not anywhere else we use GetQuantity.
    d.c_box.key:unicodeInput(true)
    game:registerDialog(d)
  elseif def.values then
    local vals = {}
    for _, val in ipairs(def.values) do
      local name = def.list_values and def.list_values[val] or val
      table.insert(vals, { name = name, value = val })
    end
    if def.values_hook then
      self:triggerHook{ def.values_hook, list=vals }
    end
    local callback = function(sel)
      if not sel or not sel.name then return end
      item.value = sel.value
      item.display_val = def.short_values and def.short_values[item.value] or item.value
      config.settings.tome2[item.key] = item.value
      c_list:drawItem(item)
      if def.on_change then def.on_change(item.value, birther) end
      if set_cb then set_cb(item.key) end
    end
    ui_util.listPopup(def.name, 'Select value:', vals, 300, callback)
  end
end

-- [static] Called from the TreeList created by createOptionsList() to
-- create the description of the provided option.
function _M:computeOptionDesc(key, birther)
  local def = self.options_def[key]
  if def.group == 'birth' and not birther then
    return birth_desc .. def.desc
  else
    return def.desc
  end
end
