-- $Id: ResetRecall.lua 3413 2018-06-13 03:53:56Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local Button = require 'engine.ui.Button'
local TextzoneList = require 'engine.ui.TextzoneList'
local Numberbox = require 'engine.ui.Numberbox'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(enforce_depth, talent_actor, callback)
  self.enforce_depth = enforce_depth
  self.talent_actor = talent_actor
  self.callback = callback

  self:generateList()
  Dialog.init(self, 'Change Zone', 1, 1)

  self.c_zones = List.new {
    width = 400,
    height = 500,
    scrollbar = true,
    list = self.list,
    fct = function(item) self:use(item) end,
  }

  self.c_lvl_desc = TextzoneList.new {
    width = 250,
    height = 2*self.font_h + 12,
    weakstore = true,
  }

  self.c_lvl = Numberbox.new {
    title = 'Level: ',
    number = 1,
    min = 1,
    max = 99,
    chars = 2,
    fct = function(text) self:finish() end,
  }

  self.c_ok = Button.new {
    text = 'OK',
    fct = function() self:finish() end,
  }
  self.c_cxl = Button.new {
    text = 'Cancel',
    fct = function() self:cancel() end,
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_zones },
    { right=0, top=0, ui=self.c_lvl_desc },
    { right=0, top=self.c_lvl_desc, ui=self.c_lvl },
    { right=0, bottom=0, ui=self.c_cxl },
    { right=self.c_cxl, bottom=0, ui=self.c_ok },
  }
  self:setFocus(self.c_zones)
  self:setupUI(true, true)
end

function _M:on_register()
  game:onTickEnd(function() self.key:unicodeInput(true) end)
end

function _M:use(item)
  if not item then return end
  self.sel_zone = item
  local text = ('%s\nLevels %d-%d'):format(item.name, item.min, item.max)
  self.c_lvl_desc:switchItem(text, item.name)
  self.c_lvl.min = item.min
  self.c_lvl.max = item.max
  self.c_lvl.number = util.bound(self.c_lvl.number, item.min, item.max)
  self:setFocus(self.c_lvl)
end

function _M:finish()
  if not self.sel_zone then return end
  local lvl = self.c_lvl.number
  if not lvl or lvl < self.sel_zone.min or lvl > self.sel_zone.max then
    return
  end

  game.state.recall_dungeon = {
    to_town = false,
    zone = item.zone,
    level = lvl - self.sel_zone.min + 1,
    name = ('level %d of %s'):format(lvl, item.name),
  }
  if self.talent_actor then self.talent_actor:talentDialogReturn(true) end
  game:unregisterDialog(self)
  if self.callback then self.callback() end
end

function _M:cancel()
  if self.talent_actor then self.talent_actor:talentDialogReturn(false) end
  game:unregisterDialog(self)
end

function _M:generateList()
  local list = {}

  -- Overkill for now, but I suppose we might eventually have addons that
  -- add new zones...
  local function parse(base, add, add_simple)
    for _, dir in ipairs(fs.list(base.."/zones/")) do
      local f = loadfile(base.."/zones/"..dir.."/zone.lua")
      if f then
	setfenv(f, setmetatable({}, {__index=_G}))
	local ok, z = pcall(f)
	-- No towns, wilderness or worldmap.
	ok = ok and not z.town and not z.wilderness and not z.worldmap
	-- Nothing that we're expressly forbidden from remembering recall
	-- depth for.
	ok = ok and not z.no_recall_depth
	-- Block NO_RECALL dungeons and Lost Temple
	ok = ok and z.short_name ~= 'lost-temple' and not z.generator.map.flags.NO_RECALL
	if self.enforce_depth then
	  -- Only zones we have recall depths for.
	  ok = ok and game.state.zone_recall_depth[z.short_name]
	end
	if ok then
	  list[#list+1] = {
	    name = z.name,
	    zone = add..dir,
	    min = z.level_range and z.level_range[1],
	    max = (self.enforce_depth and game.state.zone_recall_depth[z.short_name]) or (z.level_range and z.level_range[2]),
	  }
	end
      end
    end
  end

  parse('/data', '')
  for _, dir in ipairs(fs.list("/")) do
    local _, _, addon = dir:find("^data%-(.+)$")
    if addon then parse("/"..dir, addon.."+", addon) end
  end

  table.sort(list, function(a, b)
    return a.min < b.min or (a.min == b.min and a.name < b.name)
  end)

  local chars = {}
  for i, v in ipairs(list) do
    v.name = v.name
    chars[self:makeKeyChar(i)] = i
  end
  list.chars = chars

  self.list = list
end
