-- $Id: ui-util.lua 3479 2018-10-27 15:48:57Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Assorted UI-related utility functions.

local Dialog = require 'engine.ui.Dialog'

ui_util = {}

-- A modified version of Dialog:listPopup() that computes its own height.
-- It retains the original's usability with ActorTalents:talentDialog(), in
-- particular assuming that the provided 'fct' method will be calling
-- :talentDialogReturn()
function ui_util.listPopup(title, text, list, w, fct)
  local d = Dialog.new(title, 1, 1)
  local desc = require("engine.ui.Textzone").new {
    width = w,
    auto_height = true,
    text = text,
    scrollbar = true
  }
  local l = require("engine.ui.List").new {
    width = w,
    nb_items = #list,
    list = list,
    fct = function() d.key:triggerVirtual("ACCEPT") end
  }
  d:loadUI{
    {left = 3, top = 3, ui=desc},
    {left = 3, top = 3 + desc.h + 3, ui=require("engine.ui.Separator").new{dir="vertical", size=w - 12}},
    {left = 3, bottom = 3, ui=l},
  }
  d.key:addBind("EXIT", function()
    if fct then fct() end
    game:unregisterDialog(d)
  end)
  d.key:addBind("ACCEPT", function()
    if list[l.sel].fct then
      list[l.sel].fct(list[l.sel])
      return
    end
    if fct then fct(list[l.sel]) end
    game:unregisterDialog(d)
  end)
  d:setFocus(l)
  d:setupUI(true, true)
  game:registerDialog(d)
  return d
end

-- A modified version of Dialog:yesnoPopup(), with its exit sequence
-- modified slightly to be usefully usable with ActorTalents:talentDialog().
-- In particular, the design assumes that the provided 'fct' method will be
-- calling :talentDialogReturn().
function ui_util.yesnoPopup(title, text, fct, yes_text, no_text, no_leave, escape)
  local Button = require 'engine.ui.Button'
  local Textzone = require 'engine.ui.Textzone'
  local w, h = Dialog.font:size(text)
  local d = Dialog.new(title, 1, 1)

  local ok = Button.new{
    text=yes_text or "Yes",
    fct=function()
      fct(true)
      game:unregisterDialog(d)
    end
  }
  local cancel = Button.new{
    text = no_text or "No",
    fct = function()
      fct(false)
      game:unregisterDialog(d)
    end
  }
  if not no_leave then
    d.key:addBind("EXIT", function()
      fct(escape)
      -- ??? The original yesnoPopup() calls this twice, so we do as well.
      game:unregisterDialog(d)
      game:unregisterDialog(d)
    end)
  end
  d:loadUI{
    {left=3, top=3, Textzone.new{width=w+20, height=h+5, text=text}},
    {left=3, bottom=3, ui=ok},
    {right=3, bottom=3, ui=cancel},
  }
  d:setFocus(ok)
  d:setupUI(true, true)

  game:registerDialog(d)
  return d
end

-- A modified version of Dialog:yesnocancelPopup(), with its exit sequence
-- modified slightly to be usefully usable with ActorTalents:talentDialog().
-- In particular, the design assumes that the provided 'fct' method will be
-- calling :talentDialogReturn().
function ui_util.yesnocancelPopup(title, text, fct, yes_text, no_text, cancel_text, no_leave, escape)
  local Button = require 'engine.ui.Button'
  local Textzone = require 'engine.ui.Textzone'
  local w, h = Dialog.font:size(text)
  local d = Dialog.new(title, 1, 1)

  --	d.key:addBind("EXIT", function() game:unregisterDialog(d) fct(false) end)
  local ok = Button.new{
    text=yes_text or "Yes",
    fct=function()
      fct(true, false)
      game:unregisterDialog(d)
    end
  }
  local no = Button.new{
    text = no_text or "No",
    fct = function()
      fct(false, false)
      game:unregisterDialog(d)
    end
  }
  local cancel = Button.new{
    text = cancel_text or "Cancel",
    fct = function()
      fct(false, true)
      game:unregisterDialog(d)
    end
  }
  if not no_leave then
    d.key:addBind("EXIT", function()
      -- ??? The original yesnocancelPopup() calls this twice, so we do as well.
      fct(false, not escape)
      game:unregisterDialog(d)
      game:unregisterDialog(d)
    end)
  end
  d:loadUI{
    {left=3, top=3, ui=Textzone.new{width=w+20, height=h + 5, text=text}},
    {left=3, bottom=3, ui=ok},
    {left=3+ok.w, bottom=3, ui=no},
    {right=3, bottom=3, ui=cancel},
  }
  d:setFocus(ok)
  d:setupUI(true, true)

  game:registerDialog(d)
  return d
end

-- A modified version of Dialog:yesnoLongPopup(), with its exit sequence
-- modified slightly to be usefully usable with ActorTalents:talentDialog().
-- In particular, the design assumes that the provided 'fct' method will be
-- calling :talentDialogReturn().
function ui_util.yesnoLongPopup(title, text, w, fct, yes_text, no_text, no_leave, escape)
  local Button = require 'engine.ui.Button'
  local Textzone = require 'engine.ui.Textzone'

  local d
  local ok = Button.new{
    text=yes_text or "Yes",
    fct=function()
      fct(true)
      game:unregisterDialog(d)
    end
  }
  local cancel = Button.new{
    text=no_text or "No",
    fct=function()
      fct(false)
      game:unregisterDialog(d)
    end
  }

  w = math.max(w + 20, ok.w + cancel.w + 10)

  d = Dialog.new(title, w + 6, 1)

  if not no_leave then
    d.key:addBind("EXIT", function()
      fct(escape)
      -- ??? The original yesnoLongPopup() calls this twice, so we do as well.
      game:unregisterDialog(d)
      game:unregisterDialog(d)
    end)
  end
  d:loadUI{
    {left=3, top=3, ui=Textzone.new{width=w, auto_height=true, text=text}},
    {left=3, bottom=3, ui=ok},
    {right=3, bottom=3, ui=cancel},
  }
  d:setFocus(ok)
  d:setupUI(false, true)

  game:registerDialog(d)
  return d
end

-- Creates and returns a stub Actor based on the provided base Entity and
-- optional override fields, suitable for use as the 'npc' parameter to
-- engine.Chat:init().  'base' may be the 'define_as' of an Actor entity,
-- which will be looked up in the current zone's npc_list{}.
function ui_util.chat_npc(base, override)
  local Actor = require 'mod.class.Actor'
  override = override or {}
  if type(base) == 'string' and game.zone and game.zone.npc_list then
    base = game.zone.npc_list[base] or game.zone.npc_list.ADVENTURER
  end
  local base_color = base.color
  if not base_color and base.color_r then
    base_color = { r=base.color_r, g=base.color_g, b=base.color_b }
  end
  return Actor.new {
    name = (override.name or base.name or 'Unknown NPC'),
    display = (override.display or base.display or '?'),
    color = (override.color or base_color or colors.C_w),
    image = (override.image or base.image),
    faction = (override.faction or base.faction or 'neutrals'),
  }
end

-- A modified version of engine.Chat:invoke() that returns an instance of
-- our local TalentChat dialog designed to work with :talentDialog().
-- In particular, the design assumes that one or more of the chat object's
-- underlying .action() methods will be calling :talentDialogReturn().
function ui_util.invoke_chat(chat, id)
  if chat.npc.onChat then chat.npc:onChat() end
  if chat.player.onChat then chat.player:onChat() end

  local TalentChat = require 'mod.dialogs.TalentChat'
  local d = TalentChat.new(chat, id or chat.default_id)
  game:registerDialog(d)
  return d
end

-- Returns a modified GetQuantity dialog with its exit sequence modified
-- slightly to be usefully usable with ActorTalents:talentDialog().  In
-- particular, the design assumes that the provided 'action' method will be
-- calling :talentDialogReturn().
function ui_util.get_quantity(title, prompt, default, max, action, min)
  local GetQuantity = require 'engine.dialogs.GetQuantity'
  local d = GetQuantity.new(title, prompt, default, max, action, min)
  d.okclick = function(self)
    self.qty = self.c_box.number
    if self.qty then
      self.action(self.qty)
      game:unregisterDialog(self)
    else
      Dialog:simplePopup("Error", "Enter a quantity.")
    end
  end
  return d
end

-- Returns a modified GetText dialog with its exit sequence modified
-- slightly to be usefully usable with ActorTalents:talentDialog().  In
-- particular, the design assumes that the provided 'action' method will be
-- calling :talentDialogReturn().
function ui_util.get_text(title, text, min, max, action, cancel, absolute)
  local GetText = require 'engine.dialogs.GetText'
  local d = GetText.new(title, text, min, max, action, cancel, absolute)
  d.okclick = function(self)
    self.name = self.c_box.text
    if self.name:len() >= self.min and self.name:len() <= self.max then
      self.action(self.name)
      game:unregisterDialog(self)
    else
      Dialog:simplePopup("Error", ("Must be between %i and %i characters."):format(self.min, self.max))
    end
  end
  return d
end
